#Tony BONNAIRE
#created on October, 2018

from sklearn.neighbors import NearestNeighbors, KDTree
import numpy as np
import igraph as ig
import matplotlib.pyplot as plt
from scipy import sparse
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture
from sklearn.mixture import BayesianGaussianMixture
import time
import trex_python.utility as utility
from scipy.spatial import Delaunay, cKDTree
import scipy.sparse as sp
from tqdm import tqdm

axSize = [0.1,0.1,0.8,0.8]
cbarPos = [0.9, 0.1, 0.025, 0.8]
figSize = (10, 8)

#==========================
#  Construct network
#==========================
def construct_network_KNN(data, K_NN):
    '''Compute the K-NN algorithm using kd_tree algorithm

    [+] data: (N,D) numpy array
            N vectors of D elements (D dimensions, N datapoints)
    [+] K_NN: float
            Number of neighbors

    [-] indices:
            Vector of N elements (number of datapoints) from which element i is a vector containing
            the list of indices that are connected to i
    '''
    nbrs = NearestNeighbors(n_neighbors=K_NN, algorithm='kd_tree').fit(data)
    distances, indices = nbrs.kneighbors(data)
    return indices



def construct_network_SISG(data, k):
    '''Build a graph with SISG criteria

    [+] data: (N,D) numpy array
            N vectors of D elements (D dimensions, N datapoints)
    [+] k: float
            Density parameter

    [-] indices:
            Vector of N elements (number of datapoints) from which element i is a vector containing
            the list of indices that are connected to i
    '''
    indices = []

    if len(data.T) == 3:
        x, y, z = data.T
    else:
        x, y = data.T
        z = []

    nbrs = NearestNeighbors(n_neighbors=2, algorithm='kd_tree').fit(data)
    distances, _ = nbrs.kneighbors(data)

    distances = distances.T[1]  #Distance to the closest neighbor

    for t in range(len(x)):
        if len(z) == 0:
            distToCurrentPoint = np.sqrt((x-x[t])**2 + (y-y[t])**2)
        else:
            distToCurrentPoint = np.sqrt((x-x[t])**2 + (y-y[t])**2 + (z-z[t])**2)

        tmp = [t]
        tmp2 = np.where((distToCurrentPoint < k*distances[t]) & (distToCurrentPoint > 0))[0]
        tmp = np.hstack((tmp, tmp2))
        indices.append(tmp)

    return indices



def construct_network_fixedLength(data, R):
    '''Link vertices using a fixed length R given as a parameter

    [+] data: (N,D) numpy array
            N vectors of D elements (D dimensions, N datapoints)
    [+] R: float
            fixed radius to link datapoints (same unit as data)

    [-] indices:
        vector of N elements (number of datapoints) from which element i is a vector containing
        the list of indices that are connected to i
    '''
    if len(data.T) < 2 or len(data.T) > 3:
        raise Exception("Don't know how to work with " + str(len(data.T)) + " dimensions")

    kdtree = KDTree(data)
    indices = kdtree.query_radius(data, R)
    indices = indices.tolist()

    return indices



def construct_network_ellipsoid(data, R):
    '''Link vertices using a fixed length R given as a parameter

    [+] data: (N,D) numpy array
            N vectors of D elements (D dimensions, N datapoints)
    [+] R: float
            fixed radius to link datapoints (same unit as data)
    [-] indices:
        vector of N elements (number of datapoints) from which element i is a vector containing
        the list of indices that are connected to i
    '''
    border = 75000
    margin = 500

    if len(data.T) < 2 or len(data.T) > 3:
        raise Exception("Don't know how to work with " + str(len(data.T)) + " dimensions")

    if len(data.T) == 3:
        x, y, z = data.T
    else:
        x, y = data.T
        z = []

    indicesTmp = []
    indices = []
    for t in range(len(x)):
        # 1) Get all points in the radius R
        if len(z) == 0:
            distToCurrentPoint = np.sqrt((x-x[t])**2 + (y-y[t])**2)
        else:
            distToCurrentPoint = np.sqrt((x-x[t])**2 + (y-y[t])**2 + (z-z[t])**2)

        tmp = [t]
        tmp2 = np.where((distToCurrentPoint < R) & (distToCurrentPoint > 0))[0]
        indicesTmp = np.hstack((tmp, tmp2))     #Indices of points in the sphere of radius R

        #if we have enough points to compute a reasonable covariance ellipse
        #and if we are not too close to the border in any direction
        computablePoint = len(indicesTmp) > 15 and x[t] > margin and y[t] > margin and x[t] < border - margin and y[t] < border - margin
        if len(z) != 0:
            computablePoint = computablePoint and z[t] > margin and z[t] < border - margin

        if computablePoint:
            # 2) Perform a eig decomposition to get the shape of the covariance ellipse around the current point
            if len(z) == 0:
                mu = [x[t], y[t]]
                sigma = np.cov(x[indicesTmp], y[indicesTmp])
            else:
                mu = [x[t], y[t], z[t]]
                sigma = np.cov((x[indicesTmp], y[indicesTmp], z[indicesTmp]))

            eigvals, eigvects = np.linalg.eigh(sigma)
            radii = 2*np.sqrt(eigvals)

            # 3) Create an ellipse with volume V and link to all points in the ellipse
            if len(z) == 0:
                r = max(radii) / min(radii)
                u, v = eigvects[:, 0]
                a = R * r
                b = R / r
                if b < a:
                    exc = b
                    b = a
                    a = exc
                th = np.arctan2(v, u)
                isInEllipse = ( (x-mu[0])*np.cos(th) + (y-mu[1])*np.sin(th) )**2 / a**2 + ( (x-mu[0])*np.sin(th) - (y-mu[1])*np.cos(th) )**2 / b**2 <= 1
                tmp2 = np.where(isInEllipse == True)[0]
                indices.append(np.hstack((t, tmp2)))

            else:   #3D case
                indices.append(indicesTmp)
        else:
            indices.append(indicesTmp)

    return indices



def construct_network_ellipsoid_optimR(data, Rit, flagFig=0, verbose=1):
    '''Link vertices using a fixed length R found by optimising the diamater of the graph

    [+] data: (N,D) numpy array
            N vectors of D elements (D dimensions, N datapoints)
    [+] R: vector
            Range to look for the best fit
    [+] flagFig: int
            1 => plot diameter = f(R)    0 => no plot
    [+] verbose: int
            1 => display info of the computation      0 => nothing displayed
    '''
    diam = []
    it = 0

    if verbose:
        print("Finding optimal value of R according to the diamater of the graph")

    for R in Rit:
        it += 1
        indices = construct_network_ellipsoid(data, R)
        G, Edges = compute_network(indices)
        diam.append(G.diameter(directed=False))
        if verbose:
            print("Iteration " + str(it) + "/" + str(len(Rit)) + " -- R = " + str(utility.nice_looking_str(R)) +
              " and D = " + str(utility.nice_looking_str(diam[len(diam) - 1])))

    bestIndex = diam.index(max(diam))
    Ropt = Rit[bestIndex]
    indices = construct_network_ellipsoid(data, Ropt)

    if(flagFig):
        plt.figure()
        plt.plot(Rit, diam, 'o-')
        plt.title('Diameter of the network')
        plt.xlabel('Length')
        plt.ylabel('Diamater')

    return Ropt, max(diam), indices



def construct_network_fixedLength_optimR(data, Rit, weighted=0, flagFig=0, verbose=1):
    '''Link vertices using a fixed length R found by optimising the diamater of the graph

    [+] data: (N,D) numpy array
            N vectors of D elements (D dimensions, N datapoints)
    [+] Rit: vector
            Range to look for the best fit
    [+] flagFig: int
            1 => plot diameter = f(R)    0 => no plot
    [+] verbose: int
            1 => display info of the computation      0 => nothing displayed
    '''
    diam = []
    it = 0

    if verbose:
        print("Finding optimal value of R according to the diamater of the graph")

    for R in Rit:
        it += 1
        indices = construct_network_fixedLength(data, R)
        G, Edges = compute_network(indices)
        weight_edges = [np.linalg.norm(data[e[0]] - data[e[1]]) for e in Edges]

        if weighted == 1:
            diam.append(G.diameter(directed=False, weights=weight_edges))
        else:
            diam.append(G.diameter(directed=False))

        if verbose:
            print("Iteration " + str(it) + "/" + str(len(Rit)) + " -- R = " + str(utility.nice_looking_str(R)) +
              " and D = " + str(utility.nice_looking_str(diam[len(diam) - 1])))

    bestIndex = diam.index(max(diam))
    Ropt = Rit[bestIndex]
    indices = construct_network_fixedLength(data, Ropt)

    if(flagFig):
        plt.figure()
        plt.plot(Rit, diam, 'o-')
        plt.title('Diameter of the network')
        plt.xlabel('Length')
        plt.ylabel('Diamater')

    return Ropt, max(diam), indices



def construct_network_fixedLength_maxGC(data, Rit, flagFig=0):
    '''Link vertices using a fixed length R found by optimising the diamater of the graph

    [+] data: (N,D) numpy array
            N vectors of D elements (D dimensions, N datapoints)
    [+] Rit: vector
            Range to look for the best fit
    [+] flagFig: int
            1 => plot diameter = f(R)    0 => no plot
    '''
    it = 0
    vecN = []

    print("Finding optimal value of R according to the GC of the graph")

    for R in Rit:
        it += 1
        indices = construct_network_fixedLength(data, R)
        G, Edges = compute_network(indices)
        idx = in_giant(G)
        if idx == -1:
            n = 0
        else:
            n = np.sum(idx)
        vecN.append(n)
        print("Iteration " + str(it) + "/" + str(len(Rit)) + " -- R = " + str(utility.nice_looking_str(R)) +
              " and per_in_gc = " + str(utility.nice_looking_str(n/len(data)*100)))

    bestIndex = vecN.index(max(vecN))
    Ropt = Rit[bestIndex]
    indices = construct_network_fixedLength(data, Ropt)

    vecN = np.asarray(vecN)

    if(flagFig):
        plt.figure()
        plt.plot(Rit, vecN/len(data), 'o-')
        plt.title('Percentage of points in GC')
        plt.xlabel('Linking length')
        plt.ylabel('Percentage of points in GC')

    return Ropt, max(vecN), indices


def construct_network_gradient(data, info, K_NN1, K_NN2):
    '''First perform a nearest-neighbors algorithm with K_NN1 (> K_NN2) and then keep the K_NN2 among them with the highest info
[+] data: N vectors of D elements (D dimensions, N datapoints)
[+] info: vector of size N containing the value of interest (to maximize)
[+] K_NN1: number of intermediate neighbors in which we check the value of the info
[+] K_NN2: number of final neighbors
[-] indices: vector of N elements (number of datapoints) from which element i is a vector containing the list of indices that are connected to i'''

    nbrs = NearestNeighbors(n_neighbors=K_NN1, algorithm='kd_tree').fit(data)
    distances, indices = nbrs.kneighbors(data)

    newIndices = []
    for t in range(len(indices)):
        subset = info[indices[t][1:K_NN1]]
        indices_sorted = np.argsort(-subset)
        tmp = [t]
        for j in range(K_NN2-1):
            tmp.append(indices[t][indices_sorted[j]+1])
        newIndices.append(tmp)

    indices = newIndices

    return indices



from sklearn.utils import resample
def construct_network_min_length(data, B, perc, k):
    G_span_bs = []
    boot = []
    idxForBS = range(len(data))

    G_span, edges_span = buildMST(data)
    newG, _ = buildMST(data)
    weight_edges_span = compute_weights(data, edges_span)   #Weights of the MST

    edges_span_new = edges_span.copy()   #Copy of edges
    weight_edges_new = np.copy(weight_edges_span)  #Copy of weigts
    p = np.zeros(len(edges_span))   #Probability of edges
    edgeAlreadyCounted = np.zeros(len(edges_span_new))

    for b in range(B):
        print('Bootstrap sample ' + str(b))
        bs = np.array(resample(idxForBS, replace=False, n_samples=int(len(data)*perc)))
        boot.append(bs)
        data_b = data[bs]

#        t1 = time.time()

        #Compute the DT
#        G_DT_tmp, Edges_DT_tmp = compute_network_delaunay(data_b)

#        #Choose a graph to work with
#        G_tmp = G_DT_tmp
#        Edges_tmp = Edges_DT_tmp
#        weight_edges_tmp = compute_weights(data_b, Edges_tmp)
#
#        #Compute spanning tree
#        G_span_temp = G_tmp.spanning_tree(weights = weight_edges_tmp)
#        edges_span_tmp = G_span_temp.get_edgelist()
#
#        #Save MST and edges
#        G_span_bs.append(G_span_temp)

        #TB 200317
        G_span_temp, edges_span_tmp = buildMST_fast(data_b)
        G_span_bs.append(G_span_temp)

        #Initialized edges counted in bootstrap realization
        edgeAlreadyCounted = np.zeros(len(edges_span_new))

#        t2 = time.time()
#        print('MST took {} seconds to compute'.format(t2-t1))

        #Update the initial MST
        for (i,e) in enumerate(edges_span_tmp):
            n1 = bs[e[0]]
            n2 = bs[e[1]]       #(n1, n2) is the edge of the bs MST in the real dataset

#            tin1 = time.time()

            #if it exists, thats ok
            if (n1, n2) in edges_span_new:
                ed = edges_span_new.index((n1,n2))
                if edgeAlreadyCounted[ed] == 0:
                    p[ed] += 1/B  #We count the road that link the two nodes
                    edgeAlreadyCounted[ed] = 1

#                tin2 = time.time()
#                print('First if took {}'.format(tin2 - tin1))

            else:       #If not, we see if a close path exists or not in the tree
                sh_path = newG.get_shortest_paths(n1, n2, output='epath')
                sh_path = np.array(sh_path)
                costPath = np.sum(weight_edges_new[sh_path[0]])   #On the initial MST
                newCost = np.linalg.norm(data[n1] - data[n2])     #New cost if we add the edge

                #if the cost is not that much a gain
                if newCost > costPath/k:    #costPath < k
                    edgesOk = np.where(edgeAlreadyCounted[sh_path[0]] == 0)[0]
                    ed = sh_path[0][edgesOk]
                    p[ed] += 1/B  #We count the road that link the two nodes
                    edgeAlreadyCounted[ed] = 1

#                    tin3 = time.time()
#                    print('Not known path but ok took {}'.format(tin3 - tin1))
                    continue

                else:
                    edges_span_new.append((n1,n2))
                    p = np.append(p, 0)
                    weight_edges_new = np.append(weight_edges_new, np.linalg.norm(data[n1]-data[n2]))
                    newG = ig.Graph(edges_span_new)
                    edgeAlreadyCounted = np.append(edgeAlreadyCounted, 0)


#                    tin4 = time.time()
#                    print('Not known path to add took {}'.format(tin4 - tin1))

#        t3 = time.time()
#        print('Post-processing {} edges took {} seconds'.format(len(edges_span_tmp), t3-t2))

    return boot, G_span_bs, newG, edges_span_new, p



def getAdjacency_MST(data):
    '''
    Compute the sparse adjacency matrix for the MST of the data set

    [+] data: (N, D) numpy array
            Data set
    [-] MST: (N,N) scipy sparse csr matrix
            Adjacency matrix for the MST of the data
    '''
#    t_ini = time.time()
    tri = Delaunay(data, qhull_options="QJ")   #QJ to force all data points to be a vertex of the DT (or use Qc then tri.coplanar simplicies)
#    t2 = time.time()
#    print('----> Scipy DT took {}' .format(t2 - t_ini))

    Edges = []
    if len(data.T) == 2:
        vec = np.array([[0, 1], [0, 2], [1, 2]])
    else:
        vec = np.array([[0, 1], [0, 2], [1, 2], [0, 3], [1, 3], [2, 3]])
    for i in vec:
        stri = np.sort(tri.simplices[:, i], axis=1)
        Edges.append(stri)

    if len(tri.coplanar) > 0:
        stri = np.sort(tri.coplanar[:,[0,2]], axis=1) #np.sort(tri.coplanar[:, i], axis=1)
        Edges.append(stri)

    Edges = np.vstack(Edges)
    weight_edges = compute_weights(data, Edges)

#    A = getAdjacency_sparse(Edges, len(data), w=weight_edges)

    A = sp.csr_matrix((weight_edges, (Edges.T[0], Edges.T[1])), shape=(len(data), len(data)))
    count = sp.csr_matrix((np.ones(len(Edges.T[0])), (Edges.T[0], Edges.T[1])), shape=(len(data), len(data)))   #To know the number of times we see each edge
    A.data /= count.data  #Get the real weight matrix

    MST = sp.csgraph.minimum_spanning_tree(A)
    return MST


def getAdjacency_bootstrapMST(data, perc=0.75, B=500, display_progress_bar=False):
    '''
    Compute the sparse adjacency matrix for the bootstrap MST of the data set

    [+] data: (N, D) numpy array
            Data set of N points with dimension D
    [+] perc: float
            Percentage of the data set used for the subsamplig (N_b = int(N*perc))
    [+] B: float
            Number of bootstrap resampling
    [-] A: (N,N) scipy sparse csr matrix
            Adjacency matrix for the average MST of the data
    '''
    N = len(data)
    Nb = int(len((data))*perc)+1

    if Nb > N:
        Nb = N

    if display_progress_bar:
        pbar = tqdm(total=B, initial=0, position=0, leave=True)

    rows = []
    cols = []
    values = []
    for i in range(B):
        idx_bs = resample(np.arange(0, len(data)), replace=False, n_samples=Nb)
        dat = data[idx_bs]

        MST = getAdjacency_MST(dat)

        row, col = MST.nonzero()

        rows.append(idx_bs[row])
        cols.append(idx_bs[col])

        rows.append(idx_bs[col])   #Symmetry
        cols.append(idx_bs[row])   #Symmetry
        values.append(np.ones(len(row)*2))

        if display_progress_bar == 1:
            pbar.update(1)

    if display_progress_bar == 1:
        pbar.close()

    rows = np.hstack(rows)
    cols = np.hstack(cols)
    values = np.hstack(values)
    AA = sp.csr_matrix((values, (rows, cols)))

    AA.data = AA.data / B
    return AA


def build_bootstrap_MST(data, B=100, perc=0.75, threshold=0.35, verbose=True):
    if verbose:
        t1 = time.time()

    #Real MST
    MST = getAdjacency_MST(data)
    Edges_mst = np.array(MST.nonzero()).T
    idx = np.where(Edges_mst.T[1] < Edges_mst.T[0])[0]
    Edges_mst = np.delete(Edges_mst, idx, axis=0)

    #BS MST
    AA = getAdjacency_bootstrapMST(data, perc, B, display_progress_bar=verbose)
#    w = AA.copy()    #TB: to get weights
    AA.data[AA.data>threshold] = 1
    AA.data[AA.data<1] = 0

    allEdges = np.array(AA.nonzero()).T
#    weights = np.array(w[AA.nonzero()])[0]   #TB: to get weights
    idx = np.where(allEdges.T[1] < allEdges.T[0])[0]
    allEdges = np.delete(allEdges, idx, axis=0)
#    weights = np.delete(weights, idx)

    #Stack and remove doublons
    allEdges = np.vstack((allEdges, Edges_mst))
    allEdges = np.unique(allEdges, axis=0)

    #Back to a list
    Edges = allEdges
    src = Edges[:,0].tolist()
    trgt = Edges[:,1].tolist()
    Edges = list(zip(src, trgt))

    #Build the graph
    G = ig.Graph(Edges)

    if verbose:
        t2 = time.time()
        print('Time to compute bootstrap MST: {} secs'.format(t2-t1))

    return G, Edges




def buildMST(data, verbose=0):
    '''Compute the MST from the Delaunay Tessellation using igraph

    [+] data: (N,D) numpy array
            N vectors of D elements (D dimensions, N datapoints)

    [-] G_span: igraph object
            Describing the resulting graph
    [-] edges_span: list
            Edges in the graph
    '''

    if verbose:
        print('buildMST -- Starting...')

    ti = time.time()
    G_DT, Edges_DT = compute_network_delaunay(data)                 #Construct network with iGraph
    t2 = time.time()
    if verbose:
        print('buildMST -- DT took {}'.format(t2-ti))

#    t1 = time.time()
    weight_edges = compute_weights(data, Edges_DT)
#    t2 = time.time()
#    print('weights took {}'.format(t2-t1))

#    t1 = time.time()
    G_span = G_DT.spanning_tree(weights=weight_edges)               #Weighted MST
#    t2 = time.time()
#    print('MST took {}'.format(t2-t1))

#    t1 = time.time()
    edges_span = G_span.get_edgelist()
#    t2 = time.time()
#    print('edgelist took {}'.format(t2-t1))

    ttot = time.time()
    if verbose:
        print('buildMST -- Time to build MST = {}'.format(ttot - ti))

    return G_span, edges_span



def buildMST_fast(data, verbose=0):
    '''Compute the MST from the Delaunay Tessellation using igraph (faster than buildMST)

    [+] data: (N,D) numpy array
            N vectors of D elements (D dimensions, N datapoints)
    [+] verbose: boolean
            Displays information

    [-] G_span: igraph object
            Describing the resulting graph
    [-] edges_span: list
            Edges in the graph
    '''

    t_ini = time.time()
    if verbose:
        print('buildMST_fast -- Starting...')

    MST = getAdjacency_MST(data)

    #Build MST from adjacency matrix
    edges_span = np.array(MST.nonzero()).T

    src = edges_span[:,0].tolist()
    trgt = edges_span[:,1].tolist()
    edges_span = list(zip(src, trgt))

    G_span=ig.Graph(edges_span, directed=False)

    t4 = time.time()
    if verbose:
        print('buildMST_fast -- Total time = {}'.format(t4 - t_ini))

    return G_span, edges_span



def build_fullGraph(data):
    '''Compute the MST from the Delaunay Tessellation using igraph (faster than buildMST)

    [+] data: (N,D) numpy array
            N vectors of D elements (D dimensions, N datapoints)

    [-] G_span: igraph object
            Describing the resulting graph
    [-] edges_span: list
            Edges in the graph
    '''

    Edges = []
    idx = np.arange(0, len(data))
    for l in idx:
        for l2 in idx:
            if l2 >= l:
                break
            Edges.append((l,l2))

    G = ig.Graph(Edges, directed=False)

    return G, Edges


def build_KNN(data, k=10):
    '''Compute a KNN graph

    [+] data: (N,D) numpy array
            N vectors of D elements (D dimensions, N datapoints)
    [+] k: int
            Number of neighbors to use

    [-] G_span: igraph object
            Describing the resulting graph
    [-] edges_span: list
            Edges in the graph
    '''
    indices_KNN = construct_network_KNN(data, k)
    G_KNN, Edges_KNN = compute_network(indices_KNN)       #Construct network with iGraph

    Edges_KNN = np.vstack(Edges_KNN)
    Edges_KNN = np.unique(Edges_KNN, axis=0)   #Remove double edges

    #Back to a list
    Edges = Edges_KNN
    src = Edges[:,0].tolist()
    trgt = Edges[:,1].tolist()
    Edges = list(zip(src, trgt))

    #Build the graph
    G_KNN = ig.Graph(Edges)

    return G_KNN, Edges_KNN


def build_KrNN(data, k=15):
    '''Compute a K reciprocal NN graph

    [+] data: (N,D) numpy array
            N vectors of D elements (D dimensions, N datapoints)
    [+] k: int
            Number of neighbors to use

    [-] G_span: igraph object
            Describing the resulting graph
    [-] edges_span: list
            Edges in the graph
    '''
    indices_KNN = construct_network_KNN(data, k)
    G_KNN, Edges_KNN = compute_network(indices_KNN)       #Construct network with iGraph

    Edges_KNN = np.vstack(Edges_KNN)
    Edges_KNN = np.unique(Edges_KNN, axis=0)   #Remove double edges

    A = getAdjacency_sparse(Edges_KNN, len(data))
    A_sym = A.T + A
    A_sym[A_sym==2] = 0
    A_sym[A_sym==4] = 1

    Edges_KNN = np.array(A_sym.nonzero()).T

    #Back to a list
    Edges = Edges_KNN
    src = Edges[:,0].tolist()
    trgt = Edges[:,1].tolist()
    Edges = list(zip(src, trgt))

    #Build the graph
    G_KNN = ig.Graph(Edges)

    return G_KNN, Edges_KNN


def build_RL(data):
    '''Compute the relative neighboroud graph

    [+] data: (N,D) numpy array
            N vectors of D elements (D dimensions, N datapoints)

    [-] G_span: igraph object
            Describing the resulting graph
    [-] edges_span: list
            Edges in the graph
    '''
    G_DT, Edges_DT = compute_network_delaunay(data)                 #Construct network with iGraph
    weight_edges = compute_weights(data, Edges_DT)

    kdt = cKDTree(data)
    ed = []

    for (i,e) in enumerate(Edges_DT):
        cl0, i0 = kdt.query([data[e[0]]], len(data)-1, distance_upper_bound=weight_edges[i])
        cl1, i1 = kdt.query([data[e[1]]], len(data)-1, distance_upper_bound=weight_edges[i])
        cl0 = cl0[0][1::]; i0 = i0[0][1::]
        cl1 = cl1[0][1::]; i1 = i1[0][1::]

        idx0 = np.where(np.isinf(cl0)==False)
        cl0 = cl0[idx0]; i0 = i0[idx0]

        idx1 = np.where(np.isinf(cl1)==False)
        cl1 = cl1[idx1]; i1 = i1[idx1]

        intersec = np.where(np.in1d(i0, i1))[0]
        if len(intersec) == 0:
            ed.append(e)

        #Remove edge if there is a point closer to e[0] or e[1] than dist(e[0], e[1])
#        if cl0 >= weight_edges[i] and cl1 >= weight_edges[i]:
#            ed.append(e)

    G = ig.Graph(ed)
#    indices = G.get_adjlist()

    return G, ed


def build_FL(data, R=150):
    indices_FL = construct_network_fixedLength(data, R)
    G_FL, Edges_FL = compute_network(indices_FL)       #Construct network with iGraph

    return G_FL, Edges_FL


#import trash.homology as homo
#def buildHOPES(data, verbose=1):
#    '''Compute the MST from the Delaunay Tessellation using igraph
#[+] data: N vectors of D elements (D dimensions, N datapoints)
#[-] G_span: igraph object describing the resulting graph
#[-] edges_span: list of all edges in the graph '''
#
#    if verbose > 0:
#        t1 = time.time()
#
#    st = homo.build_alpha_complex(data)
#    edges_mst, per0d = homo.build_minimum_spanning_tree(st)
#    critical_edges, per1d = homo.critical_edges(st)
#    positive_edges, spanGap, _ = homo.positive_edges(critical_edges, per1d, method='Bootstrap', thresh=10,
#                                                  data = data, B=30, eps=1e-2, n_processes=0)   #only used if bootstrap is chosen
#
#    #Adding a condition on the length of the edge
##    weights_positive = compute_weights(data, positive_edges)
##    _, w = homo.bootstrap_MST(data, B=10)
##    w = np.hstack(w)
##    quantile = np.percentile(w, np.linspace(0,100,1000+1))[999] #it is less than 0.1% chance that a length is above this one
##    positive_edges[weights_positive<quantile]
#
#    if verbose > 2:
#        t2 = time.time()
#        print('Total time to compute HoPeS-1D: {} secs'.format(t2-t1))
#
#    edges = []
#    for e in np.vstack((edges_mst, positive_edges)):
#        edges.append((e[0], e[1]))
#
#    G = ig.Graph(edges)
#
#    return G, edges



def removeShortEdges(F, edges_span, length):
    w = compute_weights(F, edges_span)
    edgesToRemove = np.where(w < length)[0]
    while len(edgesToRemove) > 0:
        arr_edges = np.array(edges_span)
        arr_edges = np.delete(arr_edges, edgesToRemove, axis=0)
        indexF_used = np.unique(arr_edges)
        F_new = F[indexF_used, :]

        G_span, edges_span = buildMST(F_new)
        w = compute_weights(F_new, edges_span)
        edgesToRemove = np.where(w < length)[0]
        print(edgesToRemove)
        print(str(len(F) - len(F_new)) + ' nodes removed')
        F = np.copy(F_new)
        if len(F) - len(F_new) == 0:
            break

    return F, G_span, edges_span



def compute_edges(indices):
    '''Compute the list of all edges of the network from indices vectors

    [+] indices: list
            Vector of N elements (number of datapoints) from which element i is a vector containing the list of indices that are connected to i
    [-] Edges: list
            Set of all edges in the graph given as (X,Y) corresponding to a link between points indices X and Y of the dataset
    '''
    edges = []
    for i in range(len(indices)):
        edges += [(i, indices[i][k+1]) for k in range(len(indices[i])-1)]

    return edges



def compute_weights(data, edges):
    '''Compute the weights of edges from data positions

    [+] Edges: list
            Set of all edges in the graph given as (X,Y) corresponding to a link between points indices X and Y of the dataset
    [+] data: (N,D) numpy array
        N vectors of D elements (D dimensions, N datapoints)
    '''

    edges_ar = np.array(edges)
    return np.linalg.norm(data[edges_ar[:,0]] - data[edges_ar[:,1]], axis=1)



def compute_XYZ_of_edges(edges, data):
    '''Compute the XYZ positions of edges with "none" between edges anchor points

    [+] Edges: list
            Set of all edges in the graph given as (X,Y) corresponding to a link between points indices X and Y of the dataset
    [+] data: (N,D) numpy array
        N vectors of D elements (D dimensions, N datapoints)

    [-] Xe, Ye, Ze: list
            3 lists of X,Y,Z positions of beginning and ending of edges, each separated by 'None'
    '''
    wantZ = 0
    if len(data[0]) == 3:
        wantZ = 1

    Xe=[]
    Ye=[]
    Ze=[]
    for e in edges:
        Xe+=[data[e[0]][0],data[e[1]][0], None]     #x-coordinates of edge ends
        Ye+=[data[e[0]][1],data[e[1]][1], None]
        if wantZ:
            Ze+=[data[e[0]][2],data[e[1]][2], None]

    return Xe, Ye, Ze



def compute_XYZ_of_edges_noNone(edges, data):
    '''Compute the XYZ positions of edges with "none" between edges anchor points

    [+] Edges: list
            Set of all edges in the graph given as (X,Y) corresponding to a link between points indices X and Y of the dataset
    [+] data: (N,D) numpy array
        N vectors of D elements (D dimensions, N datapoints)

    [-] Xe, Ye, Ze: list
            3 lists of X,Y,Z positions of beginning and ending of edges
    '''

    wantZ = 0
    if len(data[0]) == 3:
        wantZ = 1

    Xe=[]
    Ye=[]
    Ze=[]
    for e in edges:
        Xe+=[data[e[0]][0],data[e[1]][0]]     #x-coordinates of edge ends
        Ye+=[data[e[0]][1],data[e[1]][1]]
        if wantZ:
            Ze+=[data[e[0]][2],data[e[1]][2]]

    return Xe, Ye, Ze


def compute_network(indices):
    '''Build the graph object

    [+] indices: list
            Vector of N elements (number of datapoints) from which element i is a vector containing the list of indices that are connected to i

    [-] G: igraph object
            Contains the graph structure as an igraph object
    [-] Edges: list
            Set of all edges in the graph given as (X,Y) corresponding to a link between points indices X and Y of the dataset
    '''
    Edges = []
    for i in range(len(indices)):
        Edges += [(i, indices[i][k]) for k in range(len(indices[i])) if i != indices[i][k]]

    G = ig.Graph(Edges, directed=False)

    return G, Edges


def compute_network_delaunay(data):
    '''Build the graph of the Delaunay Tesselation

    [+] indices: list
            Vector of N elements (number of datapoints) from which element i is a vector containing the list of indices that are connected to i

    [-] G: igraph object
            Contains the graph structure as an igraph object
    [-] Edges: list
            Set of all edges in the graph given as (X,Y) corresponding to a link between points indices X and Y of the dataset
    '''
    tri = Delaunay(data)

    Edges = []
    if len(data.T) == 2:
        vec = np.array([[0, 1], [0, 2], [1, 2]])
    else:
        vec = np.array([[0, 1], [0, 2], [1, 2], [0, 3], [1, 3], [2, 3]])
    for i in vec:
        stri = np.sort(tri.simplices[:, i], axis=1)
        Edges.append(stri)

    Edges = np.vstack(Edges)
    Edges = np.unique(Edges, axis=0)

    src = Edges[:,0].tolist()
    trgt = Edges[:,1].tolist()
    Edges = list(zip(src, trgt))

    G=ig.Graph(Edges, directed=False)

    return G, Edges


#==========================
#  Tools to study network
#==========================
def in_giant(G):
    '''Find the nodes in the giant component

    [+] G: igraph object
            Graph information as an igraph object

    [-] indices: list
            Indices of nodes in the giant component
    '''
    cl = G.components()
    cl_sizes = cl.sizes()
    if len(cl_sizes) != 0:
        giant_component_index = cl_sizes.index(max(cl_sizes))
    else:
        return -1
    return [x == giant_component_index for x in cl.membership]


#==========================
#  Multiscale analysis
#==========================
def iterPruning(indices, d, l_cut, edges_span):
    dIter = np.copy(d)
    n1 = len(d)
    n2 = 0
    i = 0
    init = []
    while n2 != n1:
        n1 = len(init)
        idxIter, init, _ = newPruningOnion(indices, dIter, l_cut, init)

        edges_pruned = [e for (i,e) in enumerate(edges_span) if (e[0] in idxIter and e[1] in idxIter)]
        G_pruned = ig.Graph(edges_pruned, directed=False)
        dIter = np.asarray(G_pruned.degree())
        dIter = np.append(dIter, np.zeros(len(d) - len(dIter)))
        n2 = len(init)
        i += 1
    return idxIter


def newPruningOnion(indices, d, l_cut, init=[]):
    deg = np.ones(len(d)) * d   #So that it doesn't modify the parameter d
    indExtr = np.ones(len(d), dtype='int32') * (-1)
    indBif = np.ones(len(d), dtype='int32') * (-1)

    # if we have an init we use it to set those points at -2
    if len(init) != 0:
        indExtr[init] = -2

    bifurcations = np.where(deg > 2)[0]
    extremities = np.where(deg==1)[0]
    indExtr[extremities] = extremities
    for idx in extremities:
#        print(idx)
        node = idx
        adjBif = np.array([])
        while 1:
            deg[node] -= 1
            #Check wether there are adjacent degree with deg > 0 (bifurcations)
            for (i, idNode) in enumerate(indices[node]):
                if idNode in bifurcations:
                    adjBif = [i]  #np.where(deg[indices[node]] > 2)[0]

            if len(adjBif) > 0:
                indBif[np.where(indExtr==idx)[0]] = indices[node][adjBif[0]]
                deg[indices[node][adjBif[0]]] -= 1
                break

            adjNoBif = np.where((deg[indices[node]] == 2))[0]
            if len(adjNoBif) > 0:
                pertinentId = np.where((indExtr[indices[node]] == -1) & (deg[indices[node]] == 2))[0]  #index not crossed yet
                if len(pertinentId) > 0:
                    pertinentId = pertinentId[0]
                    idAdjNoBif = indices[node][pertinentId]
                    indExtr[idAdjNoBif] = idx
            else:
                break
            deg[idAdjNoBif] -= 1    #Decrease the degree of all adjacent nodes
            node = idAdjNoBif

    idxBranches = np.unique(indExtr[indExtr>-1])
    lengthBranches = []
    for idBranch in idxBranches:
        lengthBranches.append(len(np.where(indExtr == idBranch)[0]))

    lengthBranches = np.array(lengthBranches)
    idxBranchesToKeep = np.where(lengthBranches>l_cut)[0]

    if len(idxBranchesToKeep) > 0:
        idxToKeep = np.hstack((np.where(indExtr==-1)[0], np.hstack([np.where(indExtr == i)[0] for i in idxBranches[idxBranchesToKeep]])))
    else:
        idxToKeep = np.where(indExtr==-1)[0]

    idxToThrow = np.linspace(0, len(d)-1, len(d)).astype(int)
    idxToThrow = np.delete(idxToThrow, idxToKeep)

    return idxToKeep, idxToThrow, deg


def onionDecomposition(indices, d, flagFig=1, verbose=1, edges_span=[]):
    core = 1
    currentLayer = 1
    globalLayer = 1
    deg = np.ones(len(d)) * d   #So that it doesn't modify the parameter d

    V = np.ones(len(d))
    coreness = np.zeros(len(d))
    layer = np.zeros(len(d))
    V[np.where(deg == 0)[0]] = 0

    layersCurrentCore = []
    kcores = []
    globalLayers = np.zeros(len(d))
#    edgesInLayer = []

    while np.sum(V) != 0.0 and np.sum(deg>0) != 0:
        nodesInLayer = np.asarray(np.where((deg <= core) & (deg > 0))[0])
        coreness[nodesInLayer] = core
        layer[nodesInLayer] = currentLayer

        layersCurrentCore.append(nodesInLayer)
        globalLayers[nodesInLayer] = globalLayer

        for idx in nodesInLayer:
            deg[indices[idx]] -= 1    #Decrease the degree of all adjacent nodes

        V[nodesInLayer] = 0        #Remove nodes of this layer from
        deg[nodesInLayer] = 0
        currentLayer += 1
        globalLayer += 1

        if np.sum(deg>0) != 0:
            #New core
            if min(deg[deg>0]) >= core + 1:
                core += 1 #min(deg[deg>0])
                currentLayer = 1

                kcores.append(layersCurrentCore)
                layersCurrentCore = []
        if verbose == 1:
            print('Layer ' + str(currentLayer) + ' with ' + str(len(nodesInLayer)) + ' nodes   ' + 'V = ', str(int(np.sum(V))))

    #Add the last core and layer
    if len(layersCurrentCore) != 0:
        kcores.append(layersCurrentCore)
        #Remaining points has to be put in the last layer
        idxLast = np.where(layer == 0)[0]
        layer[idxLast] = currentLayer
        globalLayers[idxLast] = currentLayer
        coreness[idxLast] = core


    fraction = []
    for k in kcores:      #For each cores
            for l in k:     #For each layer of the core
                fraction.append(len(l) / len(d))
    optLayer = -1
    for i in range(len(fraction)-5):
        if fraction[i] == fraction[i+1] and fraction[i] == fraction[i+2] and fraction[i] == fraction[i+3] and fraction[i] == fraction[i+4]:
            optLayer = i+1
            break;

    return coreness, layer, kcores, globalLayers


#==========================
#  Utility computations
#==========================
def getAdjacency(G):
    ''' Compute the adjacency matrix from a igraph object G. This is faster than the igraph routine.

    [+] G: igraph object
            Graph structure

    [-] A: numpy array
            Adjacency matrix
    '''
    n = G.vcount()
    A = np.zeros((n, n))
    edges = G.get_edgelist()
    if not G.is_directed():
        edges.extend([(v, u) for u, v in edges])

    rows, cols = zip(*edges)
    A[rows, cols] = 1
    return A



def getAdjacency_sparse(edges, N, w=[]):
    '''
    Compute the scipy sparse (csc_matrix) adjacency matrix from edge list

    [+] edges: (E,2) numpy array
        Contains indices of nodes to link
    [+] N: float
        Number of nodes in the graph
    [+] w: numpy array
        Contains the weights to put in the adjacency matrix (1 by default)

    [-] A: scipy sparse csc_matrix
        Adjacency (if w = []) or weights matrix

    Note: The graph is assumed undirected with no self loops and repetitions
          Hence, edges list reflects only half of the adjacency matrix.
    '''

    if w == []:
        halfA = sp.csc_matrix((np.ones(len(edges.T[0])), (edges.T[0], edges.T[1])), shape=(N, N))
    else:
        halfA = sp.csc_matrix((w, (edges.T[0], edges.T[1])), shape=(N,N))
#    halfA.data /= halfA.data  #Remove in case of doublons
    A = halfA + halfA.T

    return A



def getDegree(edges, N):
    '''
    Compute degrees of all N nodes in the graph

    [+] edges: (E,2) numpy array
        Contains indices of nodes to link
    [+] N: float
        Number of nodes in the graph

    [-] d: (N,) numpy array
        Contains the degree associated to each node of the graph
    '''

    A = getAdjacency_sparse(edges, N)
    d = np.array(A.sum(axis=0))[0]

    return d



#Tony BONNAIRE, created on November, 2018
import scipy.stats as sp_st
import numpy as np
from matplotlib.patches import Ellipse
import matplotlib.pyplot as plt
import warnings
import matplotlib as mpl

#Convert numeric value to .3 string to put it in title and so on
def nice_looking_str(x, n=2):
    return str(round(x*10**n)/10**n)

from scipy.stats import chi2
def get_cov_ellipse(cov, centre, ns, **kwargs):
    """
    Return a matplotlib Ellipse patch representing the covariance matrix
    cov centred at centre and scaled by the factor nstd.

    """

    # Find and sort eigenvalues and eigenvectors into descending order
    eigvals, eigvecs = np.linalg.eigh(cov)
    order = eigvals.argsort()[::-1]
    eigvals, eigvecs = eigvals[order], eigvecs[:, order]

    # The anti-clockwise angle to rotate our ellipse by
    vx, vy = eigvecs[:,0][0], eigvecs[:,0][1]
    theta = np.arctan2(vy, vx)

    # Width and height of ellipse to draw
    s = chi2(2).ppf(ns) #4.605   #95%: 5.991  90%: 4.605
    width, height = np.sqrt(eigvals*s)*2

    return mpl.patches.Ellipse(xy=centre, width=width, height=height,
                   angle=np.degrees(theta), **kwargs)


def confidence_ellipse(centre, cov, ax, n_std=3.0, facecolor='none', **kwargs):
    """
    Create a plot of the covariance confidence ellipse of `x` and `y`

    Parameters
    ----------
    x, y : array_like, shape (n, )
        Input data.

    ax : matplotlib.axes.Axes
        The axes object to draw the ellipse into.

    n_std : float
        The number of standard deviations to determine the ellipse's radiuses.

    Returns
    -------
    matplotlib.patches.Ellipse

    Other parameters
    ----------------
    kwargs : `~matplotlib.patches.Patch` properties
    """

    pearson = cov[0, 1]/np.sqrt(cov[0, 0] * cov[1, 1])
    # Using a special case to obtain the eigenvalues of this
    # two-dimensionl dataset.
    ell_radius_x = np.sqrt(1 + pearson)
    ell_radius_y = np.sqrt(1 - pearson)
    ellipse = Ellipse((0, 0),
        width=ell_radius_x * 2,
        height=ell_radius_y * 2,
        facecolor=facecolor,
        **kwargs)

    # Calculating the stdandard deviation of x from
    # the squareroot of the variance and multiplying
    # with the given number of standard deviations.
    scale_x = np.sqrt(cov[0, 0]) * n_std
    mean_x = centre[0]

    # calculating the stdandard deviation of y ...
    scale_y = np.sqrt(cov[1, 1]) * n_std
    mean_y = centre[1]

    transf = mpl.transforms.Affine2D() \
        .rotate_deg(45) \
        .scale(scale_x, scale_y) \
        .translate(mean_x, mean_y)

    ellipse.set_transform(transf + ax.transData)
    return ax.add_patch(ellipse)



def statisical_summary(x):
    return dict(zip(['Mean', 'Max', 'Min', 'Value'], [np.mean(x), max(x), min(x), x]))


def get_boundingBox(data):
    N, D = data.shape

    boundingBox = np.asarray([np.min(data.T[0]), np.max(data.T[0])])

    for d in range(1, D):
        boundingBox = np.vstack((boundingBox, np.asarray([np.min(data.T[d]), np.max(data.T[d])])))

    ranges = boundingBox[:, 1] - boundingBox[:, 0]
    shortestRange = np.min(ranges, axis=0)
    boundingBox = boundingBox.T
    return boundingBox, shortestRange


def weighted_median(data, weights):
    """
    Args:
      data (list or numpy.array): data
      weights (list or numpy.array): weights
    """
    data, weights = np.array(data).squeeze(), np.array(weights).squeeze()
#    s_data, s_weights = map(np.array, zip(*sorted(zip(data, weights))))
    s_data = data
    s_weights = weights
    midpoint = 0.5 * sum(s_weights)
    if any(weights > midpoint):
        w_median = (data[weights == np.max(weights)])[0]
    else:
        cs_weights = np.cumsum(s_weights)
        idx = np.where(cs_weights <= midpoint)[0][-1]
        if cs_weights[idx] == midpoint:
            w_median = np.mean(s_data[idx:idx+2])
        else:
            w_median = s_data[idx+1]
    return w_median


def plot_covariance_ellipse(ax, mu, sigma, p=0.9, color="k"):
    s = sp_st.chi2.ppf(p, 2)
    sigma = s*sigma

    #Compute eigenvalues/eigenvectors of the covariance matrix
    eigvals, eigvects = np.linalg.eigh(sigma)

    x, y = eigvects[:, 0]
    theta = np.degrees(np.arctan2(y, x))

    # Eigenvalues gives access to great and little axes of ellipse
    w, h = 2 * np.sqrt(eigvals)
    ellipse = Ellipse(mu, w, h, theta, color=color)
    ellipse.set_clip_box(ax.bbox)
    ellipse.set_alpha(0.5)
    ax.add_artist(ellipse)



def plot_ellipsoide(ax, mu, sigma, color="k"):
    # Find the rotation matrix and the great axis values by computing an SVD
    _, s, rot = np.linalg.svd(sigma)
    radii = np.sqrt(s)

    # Set of all spherical angles:
    u = np.linspace(0, 2 * np.pi, 10)
    v = np.linspace(0, np.pi, 10)

    # Cartesian coordinates that correspond to the spherical angles:
    # (this is the equation of an ellipsoid):
    x = radii[0] * np.outer(np.cos(u), np.sin(v))
    y = radii[1] * np.outer(np.sin(u), np.sin(v))
    z = radii[2] * np.outer(np.ones_like(u), np.cos(v))

    #Apply the rotation matrix and center the ellipsoid on mu
    for i in range(len(x)):
        for j in range(len(x)):
            [x[i,j],y[i,j],z[i,j]] = np.dot([x[i,j],y[i,j],z[i,j]], rot) + mu

    # Plot:
    ax.plot_surface(x, y, z, alpha=0.5, color=color)



def exportGraphAsTxt(file, edges):
    f = open(file, "w+")
    for e in edges:
        f.write("%d\t%d\n" %(e[0], e[1]))
    f.close()
    return


def getFigure(figSize, axSize):
    fig = plt.figure(figsize=figSize)
    ax = fig.add_axes(axSize, aspect='equal')

    return fig, ax


def importGraphAsTxt(file):
    f = open(file, "r")
    e = []
    for line in f:
        a, b = line.split()
        e.append(tuple((int(a), int(b))))
    f.close()
    return e



def density_map_NGP(posx,posy,Nx,Ny,wgt=None):
    if wgt == None:
        wgt = np.ones(len(posx))
    map_2D=np.zeros((Nx,Ny))
    npart=len(posx)
    xmin=min(posx)-1
    xmax=max(posx)+1
    ymin=min(posy)-1
    ymax=max(posy)+1

    dx=(xmax-xmin)/(1.0*Nx)
    dy=(ymax-ymin)/(1.0*Ny)

    print('xmin xmax (Mpc/h) {0} {1}'.format(xmin,xmax))
    print('ymin ymax (Mpc/h) {0} {1}'.format(ymin,ymax))
    print('dx dy (kpc/h) {0} {1}'.format(dx*1e3,dy*1e3))

    idx=(posx-xmin)/(dx)
    idy=(posy-ymin)/(dy)

    idx= np.rint(idx).astype(int)
    idy= np.rint(idy).astype(int)

    idx[idx>(Nx-1)]=Nx-1
    idx[idx<0]=0
    idy[idy>(Ny-1)]=Ny-1
    idy[idy<0]=0

    for ip in range(npart):
#          print('ip {0} / {1} test {2} , {3}'.format(ip,npart,idx[ip],idy[ip]))
         ix=idx[ip]
         iy=idy[ip]
         map_2D[ix,iy]=map_2D[ix,iy]+wgt[ip]

    map_2D=map_2D/dx/dy
    return map_2D



def plot_map(map_2D, ax=None, cbarTitle='$1+\delta$', cmap='viridis', extent=[], fig=None, cbarPos=[]):
     from matplotlib.colors import LogNorm
     if ax == None:
         fig, ax = plt.subplots()
     if len(extent)>0:
         image = ax.imshow(map_2D, origin='lower', cmap=cmap, extent=extent, aspect='auto')
         image = ax.imshow(map_2D, origin='lower', norm=LogNorm(vmin=np.min(map_2D[map_2D>0]), vmax=1), cmap=cmap, extent=extent, aspect='auto')
     else:
         image = ax.imshow(map_2D, origin='lower', cmap=cmap, aspect='auto')
         image = ax.imshow(map_2D, origin='lower', norm=LogNorm(vmin=np.min(map_2D[map_2D>0]), vmax=1), cmap=cmap, aspect='auto')

     if len(cbarPos) > 0:
         colorbar_ax = fig.add_axes(cbarPos)
         cbar = plt.colorbar(image, cax=colorbar_ax)
         cbar.set_label(cbarTitle) #, rotation=270)
     else:
         cbar = plt.colorbar(image, ax=ax)
         cbar.set_label(cbarTitle) #, rotation=270)
#         cbar.ax.get_yaxis().labelpad = 15
     return ax



def hessian(x):
    """
    Calculate the hessian matrix with finite differences
    Parameters:
       - x : ndarray
    Returns:
       an array of shape (x.dim, x.ndim) + x.shape
       where the array[i, j, ...] corresponds to the second derivative x_ij
    """
    x_grad = np.gradient(x)
    hessian = np.empty((x.ndim, x.ndim) + x.shape, dtype=x.dtype)
    for k, grad_k in enumerate(x_grad):
        # iterate over dimensions
        # apply gradient again to every component of the first derivative.
        tmp_grad = np.gradient(grad_k)
        for l, grad_kl in enumerate(tmp_grad):
            hessian[k, l, :, :] = grad_kl
    return x_grad, hessian.T


def dot(v,w):
    x,y,z = v
    X,Y,Z = w
    return x*X + y*Y + z*Z

def length(v):
    x,y,z = v
    return np.sqrt(x**2 + y**2 + z**2)

def vector(b,e):
    x,y,z = b
    X,Y,Z = e
    return (X-x, Y-y, Z-z)

def unit(v):
    x,y,z = v
    mag = length(v)
    return (x/mag, y/mag, z/mag)

def distance(p0,p1):
    return length(vector(p0,p1))

def scale(v,sc):
    x,y,z = v
    return (x * sc, y * sc, z * sc)

def add(v,w):
    x,y,z = v
    X,Y,Z = w
    return (x+X, y+Y, z+Z)

#Thanks to http://www.fundza.com/vectors/point2line/index.html
def pnt2line(pnt, start, end):
    #Not vector form
    line_vec = vector(start, end)
    pnt_vec = vector(start, pnt.T)
    line_len = length(line_vec)
    line_unitvec = unit(line_vec)
    pnt_vec_scaled = scale(pnt_vec, 1.0/line_len)
    t = dot(line_unitvec, pnt_vec_scaled)

    if np.size(t) > 1:
        t[t<0.0] = 0.0
        t[t>1.0] = 1.0
    else:
        if t < 0.0:
            t = 0.0
        elif t > 1.0:
            t = 1.0

    nearest = scale(line_vec, t)
    dist = distance(nearest, pnt_vec)
    nearest = add(nearest, start)

    return (dist, nearest)


def vectorized_pnt2line(pnt, start, end):

    #Vector form
    line_vec = end - start
    pnt_vec = pnt.T - start
    line_len = np.array([np.linalg.norm(line_vec, axis=1)]).T
    line_unitvec = line_vec / line_len
    pnt_vec_scaled = pnt_vec * 1.0 / line_len
    t = np.array([np.sum(line_unitvec*pnt_vec_scaled, axis=1)]).T

    if np.size(t) > 1:
        t[t<0.0] = 0.0
        t[t>1.0] = 1.0
    else:
        if t < 0.0:
            t = 0.0
        elif t > 1.0:
            t = 1.0

    #Vector form
    nearest = line_vec * t
    dist = np.linalg.norm(pnt_vec - nearest, axis=1)
    nearest = nearest + start

    return (dist, nearest)


def create_ranges(starts, stops, N):
    steps = (1.0/N).reshape(len(starts),1)*(stops - starts)
    interpolations = []
    for (i,w) in enumerate(N):
        vec = np.kron(np.arange(int(w)+1), steps[i]).reshape((int(w)+1, len(starts.T)))
        res = np.vstack((vec + starts[i], stops[i]))
        interpolations.append(res)
    return np.vstack(interpolations)




#For computation
import numpy as np
from scipy.spatial import cKDTree
from scipy import sparse as sp
import pandas as pd

#To handle graphs
import trex_python.networkAlgo as na
import igraph as ig

#Gestion of functions
from tqdm import tqdm
import time

#Some utility functions
import trex_python.utility as utility
import trex_python.GMM_utils as gutil
import h5py

#For preprocessing of data
from sklearn.utils import resample
from sklearn.neighbors import KDTree, NearestNeighbors

#Hiding warnings
import warnings
warnings.filterwarnings("ignore")

# ===========================================================================
# Classes
# ===========================================================================
#class mixture_:
#    def __init__(self):
#        #Parameters for gaussian clusters
#        self.F = 0          #Mean position of gaussian clusters
#        self.sig = 0        #Spatial extension of clusters (defined trough A0)
#        self.pi_k = 0       #Amplitude of Gaussian clusters
#
#        #Parameters for uniform component
#        self.alpha = 0         #Amplitude of the background noise

class regGraph_():
    def __init__(self):
        self.F = 0          #Mean position of gaussian clusters
        self.sig = 0        #Spatial extension of clusters (defined trough A0)
        self.pi_k = 0       #Amplitude of Gaussian clusters
        self.alpha = 0      #Amplitude of the background noise

        self.edges = []     #List of edges of the regularised graph
        self.responsibilities = 0   #Responsibility matrix    #DEBUG

class param_:
    #Class for set of parameters of the algorithm
    def __init__(self):
        #Parameters for parallel computation
#        self.parallel = 0       #To run in parallel
#        self.n_processes = 0    #Number of processes (used if parallel=1)

        #Parameters for denoising operation
        self.denoising = ''     #Redirect to the function to use prior to compute the topology, to remove outliers/noise in the data
        self.denoisingParameter = 0              #Assume only one parameter to denoise

        #Parameters for the graph topology
        self.topology = ''      #Redirect to the function to use to compute the topological constraint of centroids
        self.niter_update_graph = 1  #Number of iterations before updating the graph topology

        #Parameters for regularisation
        self.lam = 0            #Topological constraint
        self.lam_sig = 0        #Shape parameter of the inverse-Gamma prior distribution on variances
        self.lam_pi = 0         #Precision parameter of the Gaussian prior on weights

        #Parameters for regularization
        self.A0 = 0             #Constant used in Silverman's rule
        self.maxIter = 0        #Number max of iteration
        self.eps = 0            #For convergence analysis
        self.verbose_level = 0  #0, 1 or 2 (Minimal traces, mid or all)

        #Parameters for the bootstrap
        self.B = 0
        self.perc = 0

        #Robustness to noise and adaptative local scale
        self.covariance_update = 'fixed'      #'adaptative' or 'fixed'
        self.covariance_type = 'spherical'       #'spherical', 'diag' or 'full'
        self.minSigma = 0             #Minimum variance allowed (regularization on the covariance matrix to avoid collapse)

        #Robustness to noise
        self.background_noise = False #Wether we account for a uniform component or not
        self.domain_area = 0          #Used for the probability of the background noise

        self.ksig = 7              #Maximum sigma for udpate
        
        self.sig = 0            #Initial spatial extension of clusters (defined trough A0)
        self.pi_k = 0           #Initial amplitude of Gaussian clusters
        self.alpha = 0          #Initial amplitude of the background noise

        #Logger for traces
        self.log = ''

    def set(self, data, topology, A0=0.1, lam=1, lam_sig=5, lam_pi=1, denoisingParameter=4, B=100, perc=0.75, maxIter=10,
                              eps=1e-6, verbose=0, ksig=7, covariance_type='spherical',
                              covariance_update='fixed', minSigma=0, background_noise=False, alpha=0,
                              domain_area=0, log='', niter_update_graph=1):
#        self.parallel = parallel
        self.topology = topology
        self.niter_update_graph = niter_update_graph
        self.A0 = A0
        self.lam = lam
        self.denoisingParameter = denoisingParameter
        self.B = B
        self.perc = perc
        self.maxIter = maxIter
        self.eps = eps
#        self.n_processes = n_processes
        self.verbose_level = verbose
        self.ksig = ksig
        self.covariance_type = covariance_type
        self.minSigma = minSigma
        self.log = log
        self.domain_area = domain_area
        self.background_noise = background_noise
        self.alpha = alpha
        self.covariance_update = covariance_update
        self.lam_sig = lam_sig
        self.lam_pi = lam_pi

        self.computeBandwidth(data)  #Update variance of clusters

    def show(self):
        print('\t ========= General =========')
        print('\t logger file: {}'.format(self.log))
        print('\t verbose level: {}'.format(self.verbose_level))
#        print('\t parralel = {}'.format(self.parallel))
#        print('\t n_processes = {}'.format(self.n_processes))
        print('\n \t ========= Topology =========')
        print('\t denoising method = {}' .format(self.denoising))
        print('\t pruning level = {}'.format(self.denoisingParameter))
        print('\t topology function: {}'.format(self.topology))
        print('\t update graph: {}'.format(self.niter_update_graph))
        print('\n \t ========= Regularization =========')
        print('\t A0 = {}'.format(self.A0))
        #print('\t variance = {} (can be updated in functions)'.format(self.sig))
        print('\t lambda = {}'.format(self.lam))
        print('\t lambda_sig = {}'.format(self.lam_sig))
        print('\t lambda_pi = {}'.format(self.lam_pi))
        print('\t maximum number of iterations: {}'.format(self.maxIter))
        print('\t maximum sigma for update: {}'.format(self.ksig))
        print('\t variance type: {}'.format(self.covariance_type))
        print('\t variance update: {}'.format(self.covariance_update))
        print('\t epsilon: {}'.format(self.eps))
        print('\n \t ========= Bootstrap =========')
        print('\t Number of samples B: {}'.format(self.B))
        print('\t Sizes of samples: N*{}'.format(self.perc))
        print('\n')

    def computeBandwidth(self, data):
        '''
        Compute the variance to use for gaussian kernels using the modified Silverman's rule
        [+] data: NxD numpy array
                Matrix containing D-dimensional datapoints
        '''
        sig = np.std(data.T, axis=1)
        minsig = np.min(sig[np.nonzero(sig)])       #Minimum of std in all directions
        h = self.A0 * (1 / (len(data.T)+2)) ** (1 / (len(data.T)+4)) * len(data) ** (-1/(len(data.T)+4)) * minsig

        self.sig = h**2   #Update sigma with h**2


    def setCovariance(self, cov):
        '''
        Set the variance of gaussian clusters manually
        '''
        self.sig = cov


class filaments_:
    ''' Filament class containing information on:
        - nodes: list of nodes indices forming the ridge
        - edges: list of doublets forming edges of the filaments
        - datapoints: list of datapoints indices associated to the filament
        - type: {0:extremity to extremity, 2:extremity to bifurcation , 3:bifurcation to bifurcation}
        - length: the sum of edges weights
        - curvature: 1 - (euclidean length / geodisic length)
        - spatial_uncertainty: associated to each node
        - residuals: residual value of the projection of each associated datapoint to the ridge
        '''
    def __init__(self):
        self.nodes = []             # Set of nodes forming the ridge of the filament
        self.edges = []             # Set of edges forming the ridge of the filament
        self.datapoints = []        # Datapoints assigned with the filament
        self.type = -1              # Type de filament (bif to bif, bif to extr or extr to extr)
        self.length = 0             # Length of the filament
        self.curvature = 0          # Curvature of the filament
        self.spatial_uncertainty = [] # Spatial uncertainty associated to nodes
        self.residuals = []         # Residuals of the projection on filament of the assigned datapoints

    def summary(self):
        print(' Branch of type {} made of {} nodes and {} edges'.format(self.type, len(self.nodes), len(self.edges)))
        print(' Containing {} datapoints'.format(len(self.datapoints)))
        print(' Geodesic length = {}'.format(self.length))
        print(' Curvature = {}'.format(self.curvature))
        if len(self.datapoints) > 0:
            print(' Mean residual value = {}'.format(np.mean(self.residuals)))
    pass

# ===========================================================================
# Internal Functions
# ===========================================================================
def _residualsFilaments(X, mu, edges, fils):
    '''
    -- INTERNAL FUNCTION --
    Update the input 'fils' object with the residuals of the associated galaxies in each filament

    [+] X: NxD numpy array
            Datapoints positions
    [+] F: KxD numpy array
            Gaussian clusters means
    [+] edges: list
            Indices representing connections of gaussian clusters
    [+] sigs: Kx1 numpy array
            Scalar covariances associated to nodes
    [+] fils: list
            Regularization_GMM.filaments_ objects of detected filaments
    '''

    D = X.shape[1]

    #for each position x, project it on the filamnetary structure
    #and find the probability that it has been drawn from that
    pbar = tqdm(total=len(fils), initial=0, position=0, leave=True)
#    residuals = []
    residuals = np.ones(len(X))*(-1)

    arr_edges = np.array(edges)  #From list to numpy array to allow indexing

    for (indexFil, f) in enumerate(fils):
        f.residuals = []
        ed = arr_edges[np.array(f.edges)]
        edg = ed.T

        for (i,p) in enumerate(X[f.datapoints]):
            if D == 2:
                    dist, posNearest = utility.vectorized_pnt2line(np.hstack((p, 0)),
                                                                   np.hstack((mu[edg[0]], np.array([np.zeros(len(edg[1]))]).T)),
                                                    np.hstack((mu[edg[1]], np.array([np.zeros(len(edg[1]))]).T)))
            elif D == 3:
                    dist, posNearest = utility.vectorized_pnt2line(p, mu[edg[0]], mu[edg[1]])

            dist = np.array(dist)
            posNearest = np.array(posNearest)

            argmin = np.argmin(dist)
            d = dist[argmin]

            f.residuals.append(d)
#            residuals.append(d)
            residuals[f.datapoints] = d

        pbar.update(1)

    pbar.close()

    return np.array(residuals)



def sparse_distance_matrix(kdtree_X, X, F, max_distance):
    '''
    Compute a sparse distance matrix from a sklearn.neighbours.KDTree object between positions in X and in F

    [+] kdtree_X: sklearn.neighbours.KDTree object
            KDTree over datapoints positions
    [+] X: NxD numpy array
            Datapoint positions
    [+] F: KxD numpy array
            Gaussian clusters means
    [+] max_distance: (K,) numpy array
            Maximum distance radius around the point i in F

    [-] R: (N,K) scipy.sparse.coo_matrix
            Sparse distance matrix between X and F
    '''

    N = X.shape[0]
    K = F.shape[0]

#    t0 = time.time()
    nbrs, dist = kdtree_X.query_radius(F, max_distance, return_distance=True)
#    t1 = time.time()

    dst = np.hstack(dist)
    cols = np.repeat(np.arange(len(F)), np.array(list(map(len, nbrs))))
    rows = np.hstack(nbrs)

    R = sp.coo_matrix((dst, (rows, cols)), shape=(N, K))

#    t2 = time.time()

#    print('Time for query = {:.3f}'.format(t1-t0))
#    print('Time for post process = {:.3f}'.format(t2-t1))

    return R



#import cython_test as ct
#from scipy.special import gamma  #For laplace distribution
def _computeResponsibilityMatrix(data, F, sig, pi_k, kdtree_X, covariance_type, covariance_update, k_sig=7,
                                 background_noise=False, domain_area=0, alpha=0, Kf=0):
    '''
    -- INTERNAL FUNCTION --
    Compute the responsibility matrix using pairwise distances between elements of X
    and gaussian clusters of means F and variances sig
    Note that all distances above param.ksig*max(np.sqrt(eig(sig))) are set to 0.

    [+] data: NxD numpy array
            Datapoints positions
    [+] F: KxD numpy array
            Positions of graph nodes
    [+] sig: type and size depends on covariance_type and covariance_update
                float               if covariance_type = 'spherical' and covariance_update = 'fixed'
                (K,) numpy array    if covariance_type = 'spherical' and covariance_update = 'adaptative'
                (D,) numpy array    if covariance_type = 'diag' and covariance_update = 'fixed'
                (K,D) numpy array   if covariance_type = 'diag' and covariance_update = 'adaptative'
                (D,D) numpy array   if covariance_type = 'full' and covariance_update = 'fixed'
                (K,D,D) numpy array if covariance_type = 'full' and covariance_update = 'adpatative'
            Spatial extension of clusters, assumes a fixed and scalar variance
    [+] pi_k: scalar or (K,) numpy array
            Amplitudes of gaussian clusters
            float                   if equally_distributed = True
            (K,)                    if equally_distributed = False
    [+] kdtree_X: Scipy cKDTree object
            Input KDTree computed over the data array

    [-] R: NxK sparse csc_matrix
            The normalized responsability matrix
    '''

    D = len(data.T)
    K = len(F)
    N = len(data)

    #=================
    #Spherical covariances
    #=================
    if covariance_type == 'spherical':
#        t0 = time.time()

        if isinstance(sig, np.ndarray):    #Only if covariance_update = True
            max_dist = k_sig*np.sqrt(sig)
#            t1 = time.time()
            R = sparse_distance_matrix(kdtree_X, data, F, max_distance=max_dist)
#            t2 = time.time()
#            print('Time to compute sparse distance matrix = {:.3f}'.format(t2-t1))
        else:
            max_dist = k_sig*np.sqrt(sig)
            kdtree_F = cKDTree(F)
            R = kdtree_X.sparse_distance_matrix(kdtree_F, max_distance=max_dist, output_type='coo_matrix')

#        t1 = time.time()

        if isinstance(sig, np.ndarray):     #Only if covariance_update = True
            R.data = np.exp(-R.data**2 / sig[R.col] / 2)
            R.data = 1/(2*np.pi*sig[R.col])**(D/2)*R.data    #True Gaussian (required if background_noise=True or if sigmas are different)
        else:
#            R.data = np.exp(-R.data / sig)
#            R.data = 1/(2**D)/(np.pi)**((D-1)/2)/sig**(D/2)*gamma((D+1)/2)*R.data    #Laplace distribution
            R.data = np.exp(-R.data**2 / sig / 2)
            R.data = 1/(2*np.pi*sig)**(D/2)*R.data


        if isinstance(pi_k, np.ndarray):    #Only if covariance_update = True
            R.data = pi_k[R.col]*R.data
        else:
            R.data = pi_k * R.data

#        t2 = time.time()

        resp_background = 0
        if background_noise:
            #If background noise, then add a column for another component in the responsibility matrix
            resp_background = np.ones(len(data))*alpha/domain_area
            Rtot = sp.hstack((R, np.array([resp_background]).T))
        else:
            Rtot = R

        #Normalize responsibilities (GMM)
        sc = np.array(Rtot.sum(1).T)[0]
        R.data /= sc[R.row]

        #Normalize reponsabilites (MS)
#        sc = np.array(Rtot.sum(0).T).T[0]
#        R.data /= sc[R.col]

        if background_noise:
            resp_background /= sc

#        R.data = R.data / dist_mat.data * np.sqrt(sig)  #Laplace distribution

#        t3 = time.time()

#        print()
#        print('Time for distance: {:.3f}'.format(t1-t0))
#        print('Time for updating values: {:.3f}'.format(t2-t1))
#        print('Time for bkg and normalize: {:.3f}'.format(t3-t2))

    if covariance_type == 'full':
        resp_background = []
#        R = sp.csc_matrix((N, K))
        rows = []
        cols = []
        vals = []
        for k in range(K):
            invSig = np.linalg.inv(sig[k])
            chol_prec = np.linalg.cholesky(invSig)
            log_probs = np.sum((data @ chol_prec - F[k] @ chol_prec)**2, axis=1)
            probs = np.exp(-0.5*log_probs)
            #probs[probs < 1e-6] = 0
            probs = 1/K*1/(2*np.pi)**(D/2)/(np.linalg.det(sig[k])**0.5)*probs
            r = np.where(probs>0)[0]
            rows.append(r)
            cols.append(np.repeat(k, len(r)))
            vals.append(probs[probs>0])
#            R[:,k] = np.array([probs]).T
        rows = np.hstack(rows)
        vals = np.hstack(vals)
        cols = np.hstack(cols)
        R = sp.coo_matrix((vals, (rows, cols)), shape=(N, K))

        #Normalize responsibilities (GMM)
        sc = np.array(R.sum(1).T)[0]
        R.data /= sc[R.row]

    return R, resp_background #, R_non_norm


def _findIndexWithEntry(tab, v1, v2):
    '''
    Find all indices of element (v1, v2) in table tab:

        [+] tab: Nx2 numpy array
                Array of elements
        [+] v1: float
                Value in the first dimension
        [+] v2: float
                Value in the second dimension

        [-] idx: int
                The indices of the element [v1, v2] in the tab array
    '''

    idx = np.where((tab[:, 0] == v1) & (tab[:,1] == v2))

    return idx


def _findIndexWithEntry_list(tab, v1, v2):
    '''
    Find all indices of element (v1, v2) in table tab:

        [+] tab: list
                list of elements
        [+] v1: float
                Value in the first dimension
        [+] v2: float
                Value in the second dimension

        [-] idx: int
                The indices of the element [v1, v2] in the tab array
    '''

    idx = tab.index((v1, v2))

    return idx

# ===========================================================================
# Public Functions
# ===========================================================================

def regGMM(data, F, param, computeObj=0, display_progress_bar=0, step_update_graph=1):
    '''
    Implements the regularized GMM algorithm (Bonnaire et al., 2019) using Expectation-Maximization solver:
        handle N-D computation and uses sparse matrices to optimize RAM usage

    [+] data: NxD numpy array
            Array of D-dimensional datapoints
    [+] F: KxD numpy array
            Initial positions of clusters centers
    [+] param: Instance of 'param_' class
            The set of parameters to use
    [+] display_progress_bar=0: float
            If set to 1, a progress bar is displayed
    [+] step_update_graph: int
            Number of EM iterations before updating the graph prior
    [+] computeObj: boolean (0 or 1)
            Wether or not compute the full log-posterior (computationally costly)

    [-] F: KxD numpy array
            Updated set of clusters centers
    [-] edges_span: list of doublets
            Set of edges of the optimal regularized MST
    [-] objectives: numpy array
            Set of objective values at each iteration
            Note: costly operation
    '''

#    #=====================TEST=========================
#    if param.covariance_type == 'diag':
#        D = len(data.T)
#        if param.covariance_update == 'adaptative':
#            sig = np.repeat(sig, len(F))
#            sig = np.repeat(sig, D).reshape(len(F), D)
#        else:
#            sig = np.repeat(sig, D)
#    #==================================================

    #Get value of the parameters
    lam = param.lam
    maxT = param.maxIter
    eps = param.eps
    verbose = param.verbose_level
    covariance_type = param.covariance_type
    covariance_update = param.covariance_update

    time_start = time.time()

    COMPUTE_OBJ = computeObj

    K, D = F.shape
    min_var = param.minSigma
    alpha = param.alpha

    if isinstance(param.pi_k, np.ndarray):
        pi_k = param.pi_k
    else:
        if param.pi_k == 0:
            pi_k = (1-alpha)/K
        else:
            pi_k = param.pi_k

    if isinstance(param.sig, np.ndarray):
        lam2 = param.lam
        sig = param.sig
    else:
        lam2 = lam/param.sig
        if covariance_update == 'adaptative':
            sig = np.ones(K)*param.sig
        else:
            sig = param.sig

    #Compute graph topology from current positions of centroids
#    G_span, edges_span = na.buildMST_fast(F) #param.topology(F)
    G_span, edges_span = param.topology(F)
#    lam_saved = lam
    #lam = 0

#    w = na.compute_weights(F, edges_span)
    #Laplacian of the graph
    adj = na.getAdjacency_sparse(np.array(edges_span), K)#, w=w)
    L = sp.csgraph.laplacian(adj, normed=False)

    #Degrees
    dk = na.getDegree(np.array(edges_span), len(F))

    #Used to boost the computation of the responsability matrix R
    if covariance_type == 'spherical':
        if covariance_update == 'fixed':
            kdtree = cKDTree(data)
        elif covariance_update == 'adaptative':
            kdtree = KDTree(data)

    elif covariance_type == 'diag' and covariance_update == 'fixed':
        kdtree = cKDTree(data/np.sqrt(sig))

    #Objectives values
    mov = 1e30
    vecMov = []
    objectives = []

    t1 = 0
    t2 = 0

    #Initialise progress bar
    if display_progress_bar == 1:
        pbar = tqdm(total=maxT, initial=0, position=0, leave=True)

    t = 0
    oldF = F.copy()
    old_obj = 0

    while t < maxT: #np.linalg.norm(mov) > eps:
        if t > 0 and np.max(np.linalg.norm(F-oldF, axis=1)) < eps:
            break

        if verbose >= 1:
            print('iteration ' + str(t) + '/' + str(maxT-1))

        oldF = np.copy(F)       #Save previous result for convergence
        time_it = time.time()

        #Evaluate responsibility matrix
        t11 = time.time()
        R, R_bn = _computeResponsibilityMatrix(data, F, sig, pi_k, kdtree, covariance_type, covariance_update,
                                               background_noise=param.background_noise, domain_area=param.domain_area,
                                               alpha=alpha, k_sig=param.ksig)

        t2 += time.time() -  t11

        if verbose >= 1:
            time_R = time.time()
            print('R computed in {} seconds' .format(time_R - time_it))

        tmp = np.array(R.sum(0))
        if len(tmp.shape) > 1:
            tmp = tmp[0]

        tmp[tmp<eps] = eps
        Lambda = sp.diags(tmp)
        tmp /= sig
        Lambda2 = sp.diags(tmp)

        if verbose >= 1:
            time_Lam = time.time()
            print('Lambda computed in {} seconds' .format(time_Lam - time_R))

        t0 = time.time()
        #Update positions of projected points
        if covariance_update == 'adaptative':
            A2 = L*lam2 + Lambda2
            V = R.copy()
            V.data = V.data / sig[V.col]

            F, sig0, pi_k, alpha = gutil.M_step(data, V, R, A2, Lambda, update_pos_only=False,
                                                    background_noise=param.background_noise, resp_background=R_bn)

            lam4 = param.lam_pi
            pi_k = (pi_k + lam4 * (1-alpha) / len(F)) / (1+lam4)

            #Inverse-Gamma prior on sigma
            lam3 = param.lam_sig #/param.sig
            nk = Lambda.data[0]
            aa = adj.copy()
            aa.data[aa.data>0] = 1
            dk = np.sum(aa, axis=0).A[0]
            sig_nbrs = (aa @ sig) / dk[np.newaxis, :]
            sig = np.array((sig0 + 4 * lam3 * sig_nbrs / D / nk) / (1 + 4*lam3/D/nk))[0]

            sig[sig < min_var] = min_var            #Regularisation on covariances to avoid collapse to 0 locally

        elif covariance_update == 'fixed':
            A2 = L*lam2 + Lambda2
            V = R.copy()
            V.data = V.data / sig
            F = gutil.updatePositionsGMM(data, V, A2)

            #Equivalent scalar solve of the M-step for F (slower)
#            deg = na.getDegree(np.array(edges_span), len(F))
#            ff = []
#            Rcsc = R.tocsc()
#            for k in range(len(F)):
#                aa = np.array(adj[k,:].todense())[0]
#                reg = aa.T@F
#                num = Rcsc[:,k].T@data/sig + 2*lam2/2*reg
#                den = Lambda.data[0][k]/sig + 2*lam2/2*deg[k] #- 2*lam2/2*reg#- 2*lam/2*deg[k]
#                ff.append(num[0]/den)
#
#            ff = np.array(ff)
#            F = ff.copy()
#            print('')
#            print(F)
#            print('---------------')
#            print(ff)
#            print(np.allclose(F, ff, 1e-30))

            if param.background_noise:
                alpha = 1/len(data) * np.sum(R_bn)
                pi_k = (1-alpha)/K
            
        if verbose >= 1:
            time_update = time.time()
            print('New positions computed in {} seconds'.format(time_update - time_Lam))

        if verbose >= 1:
            print('New F computed in {} seconds'.format(time.time() - time_it))

        t1 += time.time() - t0
        
#        if t < maxT:
        if t%step_update_graph == 0 and t > 0 or t<=3:
            #Put lambda to its value
            #lam = lam_saved
            
            #Compute the new graph
            G_span, edges_span = param.topology(F)

#            w = na.compute_weights(F, edges_span)
            #Laplacian of the graph
            adj = na.getAdjacency_sparse(np.array(edges_span), K)#, w=w)
            L = sp.csgraph.laplacian(adj, normed=False)

            #Degrees
            dk = na.getDegree(np.array(edges_span), len(F))

            if verbose >= 1:
                time_topo = time.time()
                print('Topology computed in {} seconds'.format(time_topo - time_update))

        t+=1

        if display_progress_bar == 1:
            #Update progress bar
            pbar.update(1)

        if COMPUTE_OBJ:
            #Compute the objective function
            obj = gutil.evaluateObjective(data, F, V, Lambda2, L, lam2, sig, pi_k,
                                          background_noise=param.background_noise, alpha=alpha, domain_area=param.domain_area)
            objectives.append(obj)

        mov = np.linalg.norm(F-oldF)
        vecMov.append(mov)
        if verbose >= 1:
            print('Total displacement = ' + str(mov))

        if mov < eps:
            #If convergence is reached, break the loop
            print('Converged: Displacement')
            break

        if t > 0 and COMPUTE_OBJ:
            if np.abs(obj - old_obj) < eps:
                #If convergence is reached, break the loop
                print('Converged: Likelihood is not moving')
                break

            old_obj = obj

    if verbose >= 1:
        print('T-ReX done! Time elapsed: {} seconds'.format(time.time()-time_start))

    vecMov = np.asarray(vecMov)
    if COMPUTE_OBJ==0:
        objectives = vecMov

    if display_progress_bar == 1:
        pbar.close()

    #Write final estimations into regGraph object
    regGraph = regGraph_()
    regGraph.sig = sig
    regGraph.alpha = alpha
    regGraph.pi_k = pi_k
    regGraph.F = F
    regGraph.edges = edges_span
    regGraph.responsibilities = (R, R_bn)

    print('Time in E-step: {:.3f}'.format(t2))
    print('Time in M-step: {:.3f}'.format(t1))

    return regGraph, objectives, vecMov, adj


def bootstrap(data, param, display_progress_bar=0):
    '''
    Computes param.B realisations of regularised GMM with param.perc*N randomly chosen datapoints

    [+] data: NxD numpy array
            Array of D-dimensional datapoints
    [+] param: Instance of 'param_' class
            The set of parameters to use
    [+] display_progress_bar=0: float
            If set to 1, a progress bar is displayed

    [-] allF: list
            Contains all numpy arrays of centroids positions of the param.B realizations
    [-] allEdges: list
            Contains all lists of edges of the param.B realizations
    '''

    #Get parameters
    B = param.B
    perc = param.perc
    verbose = param.verbose_level

    if verbose >= 1:
        print('\n Computing bootstrap with current parameters:')
        param.show()

    #Initialization
    idxForBS = range(len(data))
    allF = []
    allEdges = []

    if len(data.T) == 2:
        data_temp = np.vstack((data.T, np.zeros(len(data))))
        data_temp = data_temp.T
    else:
        data_temp = np.copy(data)

    Nb = int(len(data)*perc)

    #Initialise progress bar
    if display_progress_bar == 1:
        pbar = tqdm(total=B, initial=0, position=0, leave=True)

    for b in range(B):
        boot = resample(idxForBS, replace=False, n_samples=Nb)
        boot = np.asarray(boot)

        F, opt_edges = getRegularisedGraph(data, boot, param)

        allF.append(F)
        allEdges.append(opt_edges)

        #Update progress bar
        if display_progress_bar == 1:
            pbar.update(1)

#        if(int(b/B*100)%10 == 0 and int(b/B*100) != lastperc):
#            print(str(int(b/B*100)) + '% completed ({} seconds passed)' .format(time.time()-t_ini))
#            lastperc = int(b/B*100)

    if display_progress_bar == 1:
        pbar.close()

#    if verbose >= 0:
#        print('Bootstrap done! Time elapsed: {} seconds'.format(time.time()-t_ini))

    return allF, allEdges



def pruning(G, param):
    '''
    Compute the pruning of the graph G by iteratively removing nodes standing at branches extremities.
    The graph is pruned at a level param.denoisingParameter.

    [+] G: igraph object
            The graph built with nodes F
    [+] param: Instance of 'param_' class
            The set of parameters to use

    [-] idx: numpy array
            Indices of the nodes to keep to get the pruned graph
    '''
    l_cut = param.denoisingParameter

    if l_cut >= 0:
        indices_span = G.get_adjlist()
        d_span = np.asarray(G.degree())

#        t1 = time.time()
        coreness, layers, kcores, globalLayers = na.onionDecomposition(indices_span, d_span, 0, 0)    #no figures and verbose
#        print('Onion in {}'.format(time.time() - t1))

        idx = np.where(globalLayers > l_cut)[0]   #layers

    return idx



def getRegularisedGraph(data, idxBoot, param):
    '''
    Compute the regularised graph through regGMM function using only data[idxBoot] datapoints:
        1) Compute graph topology
        2) Prune the graph
        3) Update the bandwidth
        4) Call regGMM to get a smooth version of the pruned graph

    [+] data: NxD numpy array
            Array of D-dimensional datapoints
    [+] idxBoot: Mx1 numpy array
            Array of M (<=N) indices from data to use for the computation
    [+] param: Instance of 'param_' class
            The set of parameters to use

    [-] F: KxD numpy array
            Updated set of clusters centers
    [-] opt_edges: list of doublets
            Set of edges of the optimal regularized MST
    '''
    t_ini = time.time()

    #Get parameters for the computation
    data_b = data[idxBoot]
    verbose = param.verbose_level

    #Compute the graph topology
    G_temp, _ = param.topology(data_b)

    if verbose >= 2:
        t_mst = time.time()
        print('Time for computing graph: {}' .format(t_mst-t_ini))

    #Pruning MST
    if param.denoisingParameter > 0:
        idx_temp = pruning(G_temp, param)
    else:
        idx_temp = np.arange(0, len(data_b))

    if verbose >= 2:
        t_pruning = time.time()
        print('Time for pruning: {}' .format(t_pruning-t_mst))

    #Find smoothed skeleton of the BT sample
    X_skel = np.copy(data_b[idx_temp])
#    F_skel = np.copy(data_b[idx_temp])
    F_skel = np.copy(X_skel)
#    F_skel = X_skel[np.sort(np.random.choice(len(X_skel), 5000, replace=False))] #[idx]

    #Compute the variance sigma for clusters' extension
#    param.computeBandwidth(data=data)

    #Regularized graph
    F = []
    opt_edges = []
    try:
        regGraph, _, _, _ = regGMM(X_skel, F_skel, param, display_progress_bar=0)
        F = regGraph.F
        opt_edges = regGraph.edges
    except Exception as e:
        print(e)

    if verbose >= 2:
        t_reg = time.time()
        print('Time for regularization: {}' .format(t_reg-t_pruning))

    return F, opt_edges



def probabilityMap(data, param, centroids, edges, Nx, bins=[]):
    '''
    Build the probability map with a Nx resolution from datapoints data and computed given centroids and edges
    Edges are part of the map and must be sampled before stacking histograms

    [+] data: NxD numpy array
            Array of D-dimensional datapoints
    [+] param: Instance of 'param_' class
            The set of parameters to use
    [+] centroids: NbxD numpy array
            Set of B centroids positions computed during the bootstrap
    [+] edges: list
            List of B lists of edges obtained during each realization
    [+] Nx: dx1 numpy array
             Vector containing resolution of the resulting probability map in each dimension
    [+] bins: list of numpy arrays (OPTIONAL)
            Predefined set of bins on which to compute the map

    [-] probMap: list
            Element 0 of the list is the Nx[0]xNx[1]x...xNx[D] numpy array of the D dimensional histogram
            Element 1 of the list is the corresponding bins
    '''

    N, D = data.shape   #Get shape of data
    boundingBox, sizeBox = utility.get_boundingBox(data)
    sampling = (sizeBox/max(Nx))

    #Build the grid for histograms if not already given
    if len(bins) == 0:
        bins = []
        for d in range(D):
            bins.append(np.mgrid[min(data.T[d]):max(data.T[d]):(max(data.T[d])-min(data.T[d]))/Nx[d]])

    if param.verbose_level >= 0:
        print('Computing binary histograms')
        t1 = time.time()

    i = 0
    temp = np.histogramdd(centroids[0], bins=bins)[0]  #To get the shape
    hh = np.zeros(temp.shape)       #Initialize the hh
    for (i, ed) in enumerate(edges):
        interpolateF = centroids[i]
        test = np.array(ed)
        spaces = np.linalg.norm(interpolateF[test[:,0]]-interpolateF[test[:,1]], axis=1) / sampling * 5
        interpolateF = utility.create_ranges(interpolateF[test[:,0]], interpolateF[test[:,1]], spaces)

        temp = np.histogramdd(interpolateF, bins=bins)[0]
        temp[temp>0]=1      #binary histogram
        hh += temp/len(centroids)

        if param.verbose_level >= 0:
            if int(i/len(edges)*100)%10 == 0:
                t2 = time.time()
                print(str(int(i/len(edges)*100)) + '% done (' + str(t2-t1) + 'secs passed)')

    probMap = (hh, bins)

    return probMap



def filamentFinding(data, regGraph, param, key='weight', type='all'):
    ''' Filament identification for a graph G with a tree structure assumed. Three types of filaments exists:
               type 1 : extremity to extremity
               type 2 : extremity to bifurcation
               type 3 : bifurcation to bifurcation

    [+] data: NxD numpy array
            Array of D-dimensional datapoints
    [+] regGraph: regGraph_ object
            Contains regularized graph information (model parameters)
    [+] param: Instance of 'param_' class
            The set of parameters to use
    [+] key: String
            key for weights to use in the igraph object
    [+] type: String 'all' or anything else
            Type of filaments  o extract ('all' => all; anything else => 1 and 2 only)

    [-] filaments: list
            Set of filaments_ class instances
    '''

    t0 = time.time()

    F = regGraph.F
    edges = regGraph.edges

#    N = len(data)
    K = len(F)

    #Build an igraph object from edges and nodes
    G = ig.Graph(edges, directed=False)
    d = np.asarray(G.degree())
    d = np.append(d, np.zeros(len(F.T[0]) - len(d)))
    w = na.compute_weights(F, edges)
    G.es.set_attribute_values('weight', w)

    verbose = param.verbose_level
    filaments = []
    indices = G.get_adjlist()
    edges_array = np.array(edges)
    d = np.asarray(G.degree())
    affil_clusters = np.zeros(len(F))
    alreadyUsed = np.zeros(len(F))

    dok = sp.csc_matrix((np.arange(0, len(edges_array)), (edges_array[:,0], edges_array[:,1])), shape=(K,K)).todok()

    try:
        w = np.asarray(G.es.get_attribute_values(key))
    except:
        w = np.ones(len(edges))

    #Step 1: find the topological class of each nodes (Junction, bifurcation or extremity)
    node_topo = np.zeros(len(d))
    node_topo[d==1] = 1  #Extremity
    node_topo[d==2] = 2  #Junction
    node_topo[d>2] = 3  #Bifurcation

    extremities = np.where(node_topo == 1)[0]
    bifurcations = np.where(node_topo == 3)[0]

    # 1 single loop, exceptional case
    if len(extremities) == 0 and len(bifurcations) == 0:
        print('\n Exceptional case')
        junctions = np.where(node_topo == 2)[0]
        filament = filaments_()
        length = 0
        prev_node = -1

        j = junctions[0]     #Take a node
        filament.nodes.append(j)
        jIni = j
        jNext = indices[j][0]
        try:
            idx_edge = edges.index((jNext, j))
        except:
            idx_edge = edges.index((j, jNext))

        length = w[idx_edge]
        filament.edges.append(idx_edge) #edges[idx_edge])
        prev_node = j
        j = jNext

        while j != jIni:
            filament.nodes.append(j)
            idx = np.where(np.asarray(indices[j]) != prev_node)[0][0]
            jNext = indices[j][idx]
            try:
                idx_edge = edges.index((jNext, j))
            except:
                idx_edge = edges.index((j, jNext))
            length += w[idx_edge]
            filament.edges.append(idx_edge) #edges[idx_edge])

            prev_node = j
            j = jNext

        filament.nodes = np.array(filament.nodes)
        filament.type = 0
        filament.length = length
        filament.curvature = 1 - np.linalg.norm(F[filament.nodes[0]] - F[filament.nodes[len(filament.nodes)-1]])/length
        filament.spatial_uncertainty = np.zeros(len(filament.nodes))
        filaments.append(filament)
        return filaments

    if verbose > 1:
        t1 = time.time()
        print('\nfilamentFinding -- step 1 executed in {:.2f} secs'.format(t1 - t0))

    #Step 2: find filaments defines as segments of different types:
    #           type 1 : extremity to extremity
    #           type 2 : extremity to bifurcation
    #           type 3 : bifurcation to bifurcation

    #First looking for type 1 and 2
    for e in extremities:
        filament = filaments_()
        filament.nodes.append(e)
        i = indices[e][0]     #there should be only 1 element for extremities
        length = 0
        prev_node = e

        #TEST
        alreadyUsed[e] = 1

        while node_topo[i] != 1 and node_topo[i] != 3: #(not(i in bifurcations) and not(i in extremities)):
            filament.nodes.append(i)
#            idx_edge = _findIndexWithEntry(edges_array, min(prev_node, i), max(prev_node, i))[0][0]
            idx_edge = dok[min(prev_node, i), max(prev_node, i)]   #TB 200316: faster

            #TEST
            alreadyUsed[i] = 1

            length += w[idx_edge]
            idx = np.where(np.asarray(indices[i]) != prev_node)[0][0]
            prev_node = i       #Previous node
            i = indices[i][idx]  #There should be 2 elements for junction so we take the node we don't come from
            filament.edges.append(idx_edge) #edges[idx_edge])


        #Adding the last node (bifurcation or extremity)
#        idx_edge = _findIndexWithEntry(edges_array, min(prev_node, i), max(prev_node, i))[0][0]
        idx_edge = dok[min(prev_node, i), max(prev_node, i)]   #TB 200316: faster
        length += w[idx_edge]
        filament.nodes.append(i)
        filament.edges.append(idx_edge) #edges[idx_edge])
        prev_node = i

        typeOfFil = -1
        if node_topo[prev_node] == 1: #prev_node in extremities:
            typeOfFil = 1
        elif node_topo[prev_node] == 3: #prev_node in bifurcations:
            typeOfFil = 2

        #Save the filament of type 1 or 2 (respectively ext to ext and ext to bif)
        filament.nodes = np.array(filament.nodes)
        filament.type = typeOfFil
        filament.length = length
        filament.curvature = 1 - np.linalg.norm(F[filament.nodes[0]] - F[filament.nodes[len(filament.nodes)-1]])/length
        filament.spatial_uncertainty = np.zeros(len(filament.nodes))

        affil_clusters[filament.nodes] = len(filaments)
        filaments.append(filament)

    if verbose>1:
        t2 = time.time()
        print('filamentFinding -- step 2 executed in {:.2f} secs'.format(t2 - t1))
        print('\t' + str(len(filaments)) + ' filaments of type 1 or 2 found')

    if type == 'all':
        #Then looking for type 3 (bif to bif)
        cnt = 0
        for b in bifurcations:
            for p in indices[b]:        #For all possible path at the bifurcation

                if node_topo[p] != 3 and alreadyUsed[p]:
                    continue;   #If the point is not a bifurcation but already used, then this path is already encoded

                filament = filaments_()
                filament.nodes.append(b)
                length = 0
                prev_node = b
                alreadyUsed[p] = 1

                i = p
                while node_topo[i] != 1 and node_topo[i] != 3: #(not(i in bifurcations) and not(i in extremities)):
                    filament.nodes.append(i)
#                    idx_edge = _findIndexWithEntry(edges_array, min(prev_node, i), max(prev_node, i))[0][0]
                    idx_edge = dok[min(prev_node, i), max(prev_node, i)]   #TB 200316: faster

                    length += w[idx_edge]
                    idx = np.where(np.asarray(indices[i]) != prev_node)[0][0]
                    prev_node = i       #Previous node
                    i = indices[i][idx]  #There should be 2 elements for junction so we take the node we don't come from
                    filament.edges.append(idx_edge) #edges[idx_edge])
                    alreadyUsed[i] = 1

                #Adding the last node (bifurcations or extremities)
#                idx_edge = _findIndexWithEntry(edges_array, min(prev_node, i), max(prev_node, i))[0][0]
                idx_edge = dok[min(prev_node, i), max(prev_node, i)]   #TB 200316: faster
                length += w[idx_edge]
                filament.nodes.append(i)
                filament.edges.append(idx_edge) #edges[idx_edge])
                prev_node = i

                typeOfFil = -1
                if node_topo[prev_node] == 3: #prev_node in bifurcations:       #Save it as a filament only if it ends on a bifurcation
                    filament.type = 3
                    filament.length = length
                    filament.curvature = 1 - np.linalg.norm(F[filament.nodes[0]] - F[filament.nodes[len(filament.nodes)-1]])/length
                    filament.spatial_uncertainty = np.zeros(len(filament.nodes))
                    affil_clusters[filament.nodes] = len(filaments)
                    filaments.append(filament)
                    cnt += 1
        if verbose>1:
            print('\t' + str(cnt)  + ' filaments of type 3 found')
            t3 = time.time()
            print('filamentFinding -- Step 3 executed in {:.2f} secs'.format(t3 - t2))

    dataSeg = np.ones(len(data))*(-1)

    if verbose > 1:
        print('\nfilamentFinding -- Starting data segmentation')

    #Get the segmentation of datapoints in the different branches
    if param.covariance_type == 'spherical':
        if param.covariance_update == 'adaptative':
            kdtree = KDTree(data)
        elif param.covariance_update == 'fixed':
            kdtree = cKDTree(data)
    elif param.covariance_type == 'diag':
        kdtree = cKDTree(data/np.sqrt(param.sig))

    R, R_bn = _computeResponsibilityMatrix(data, F, regGraph.sig, regGraph.pi_k, kdtree, param.covariance_type, param.covariance_update, k_sig=7,
                                     background_noise=param.background_noise, domain_area=param.domain_area, alpha=regGraph.alpha)
#    if param.background_noise:
#        R = sp.hstack((R, np.array([R_bn]).T))

    zi = R.argmax(axis=1)
    if len(zi.shape)>1:
#        zi[zi == len(F)+1] = -1                    #Remove datapoints associated to background noise
        zi[np.where(R.sum(axis=1) == 0)[0]] = -1    #Datapoints that are above k sigma for all clusters are removed
        zi = np.array(zi.T)[0]
    else:
#        zi[zi == len(F)+1] = -1                    #Remove datapoints associated to background noise
        zi[R.sum(axis=1) == 0] = -1                 #Datapoints that are above k sigma for all clusters are removed

    #If background component dominates, reject it
#    zi[R_bn > 0.5] = -1
    zi[R_bn > np.array(R.sum(axis=1)).T[0]] = -1

    if verbose > 1:
        t4 = time.time()
        print('filamentFinding -- responsibilities computed in {:.2f} secs'.format(t4-t3))

    #zi is the affiliation of datapoints to Gaussian clusters
    #Since we have the affiliation of clusters to filament, we can link both information
    #to get the affiliation of datapoints to filaments
    for (indexFil, f) in enumerate(filaments):
        #Return indices of zi that have a non-null intersection with nodes of the considered filaments
        idx = np.argwhere(np.in1d(zi, f.nodes)).T[0]
        dataSeg[idx] = indexFil
        f.datapoints = idx

    dataSeg.astype(int)

    if verbose > 1:
        t5 = time.time()
        print('\nfilamentFinding -- Tagged assignations in {:.2f} secs'.format(t5 - t4))
        print('filamentFinding -- Computing residuals...')

    res = _residualsFilaments(data, F, edges, filaments)

    if verbose > 1:
        t6 = time.time()
#        print('\n FilamentFinding -- Residuals computed in {:.2f} secs'.format(t6 - t5))

    if verbose > 1:
        print('\nfilamentFinding -- Exit after {:.2f} secs'.format(t6 - t0))

    return filaments, dataSeg, affil_clusters, res



def assign_labels(data, regGraph, filaments, param):
    '''
    Assign a label to a set of input data according a graph and a set of extracted branches

    [+] data: NxD numpy array
            Array of D-dimensional datapoints
    [+] regGraph: regGraph_ object
            Contains regularized graph information (model parameters)
    [+] filaments: list
        Set of filaments_ class instances
    [+] param: Instance of 'param_' class
            The set of parameters to use

    [-] labels: (N,) numpy array
            Contains labels (index of filament) assigned to the ith input data
    '''

    F = regGraph.F

    if param.covariance_type == 'spherical':
        if param.covariance_update == 'adaptative':
            kdtree = KDTree(data)
        elif param.covariance_update == 'fixed':
            kdtree = cKDTree(data)

    R, R_bn = _computeResponsibilityMatrix(data, F, regGraph.sig, regGraph.pi_k, kdtree,
                                           param.covariance_type, param.covariance_update, k_sig=3,
                                           background_noise=param.background_noise,
                                           domain_area=param.domain_area, alpha=regGraph.alpha)
#    if param.background_noise:
#        R = sp.hstack((R, np.array([R_bn]).T))

    zi = R.argmax(axis=1)
    if len(zi.shape)>1:
#        zi[zi == len(F)+1] = -1                    #Remove datapoints associated to background noise
        zi[np.where(R.sum(axis=1) == 0)[0]] = -1    #Datapoints that are above k sigma for all clusters are removed
        zi = np.array(zi.T)[0]
    else:
#        zi[zi == len(F)+1] = -1                    #Remove datapoints associated to background noise
        zi[R.sum(axis=1) == 0] = -1                 #Datapoints that are above k sigma for all clusters are removed

    #If background component dominates, reject it
#    zi[R_bn > 0.5] = -1
    zi[R_bn > np.array(R.sum(axis=1)).T[0]] = -1

    #Create label vector
    labels = np.ones(len(data))*(-1)

    for (indexFil, f) in enumerate(filaments):
        #Return indices of zi that have a non-null intersection with nodes of the considered filaments
        idx = np.argwhere(np.in1d(zi, f.nodes)).T[0]
        labels[idx] = indexFil

    return labels.astype(int)


def in_out(data, regGraph, param):
    '''
    '''

    F = regGraph.F

    if param.covariance_type == 'spherical':
        if param.covariance_update == 'adaptative':
            kdtree = KDTree(data)
        elif param.covariance_update == 'fixed':
            kdtree = cKDTree(data)

    R, R_bn = _computeResponsibilityMatrix(data, F, regGraph.sig, regGraph.pi_k, kdtree,
                                           param.covariance_type, param.covariance_update, k_sig=7,
                                           background_noise=param.background_noise, domain_area=param.domain_area,
                                           alpha=regGraph.alpha)
#    if param.background_noise:
#        R = sp.hstack((R, np.array([R_bn]).T))

    zi = R.argmax(axis=1)
    if len(zi.shape)>1:
#        zi[zi == len(F)+1] = -1                    #Remove datapoints associated to background noise
        zi[np.where(R.sum(axis=1) == 0)[0]] = -1    #Datapoints that are above k sigma for all clusters are removed
        zi = np.array(zi.T)[0]
    else:
#        zi[zi == len(F)+1] = -1                    #Remove datapoints associated to background noise
        zi[R.sum(axis=1) == 0] = -1                 #Datapoints that are above k sigma for all clusters are removed

    #If background component dominates, reject it
#    zi[R_bn > 0.5] = -1
    zi[R_bn > np.array(R.sum(axis=1)).T[0]] = -1

    labels = np.ones(len(zi))
    labels[zi==-1] = 0

    return labels.astype(int)


def graphRefinement(data, regGraph, b=0, verbose=1):
    '''
    Perform a refinement of the graph based on 2 assumptions:
        - Branches of the graph structure should not have a length lower than b
        ### NOT DONE ### - Bifurcations should be merged if the path linking the 2 is straight (no branches)
                and with a length lower than b

    [+] data: (N, D) numpy array
            Array of D-dimensional datapoints
    [+] regGraph: regGraph_ object
            Contains regularized graph information (model parameters)
    [+] b: float
            Minimum value to keep branches of the graph


    [-] F_refined: (K', D) numpy array with K' <= K
            Positions of refined graph nodes
    [-] idx_refined: (K',) numpy array
            Indices in F such that F[idx_refined] = F_refined
    '''

    t0 = time.time()
    D = data.shape[1]

    F = regGraph.F
    edges = regGraph.edges

    key = 'weight'

    #Build an igraph object from edges and nodes
    G = ig.Graph(edges, directed=False)
    d = np.asarray(G.degree())
    d = np.append(d, np.zeros(len(F.T[0]) - len(d)))
    w = na.compute_weights(F, edges)
    G.es.set_attribute_values(key, w)

    if verbose >= 1:
        print()
        print('graphRefinement -- Branches shorter than {:.3f} will be removed iteratively'.format(b))

    newF = np.copy(F)

    cnt = 0
    while 'There are still branches with length < b':
        cnt += 1
        #Extract branches
        nodes_to_throw = []
        #filaments = []
        indices = G.get_adjlist()    #Could be done without igraph
        edges = G.get_edgelist()     #Could be given as an argument
        edges_array = np.array(edges)
        d = np.asarray(G.degree())   #Could be computed without igraph

        K = G.vcount()
        dok = sp.csc_matrix((np.arange(0, len(edges_array)), (edges_array[:,0], edges_array[:,1])), shape=(K, K)).todok()

        new_edges = np.vstack(edges)

        try:
            w = np.asarray(G.es.get_attribute_values(key))
        except:
            w = np.ones(len(edges))

        #Step 1: find the topological class of each nodes (Junction, bifurcation or extremity)
        node_topo = np.zeros(len(d))
        node_topo[d==1] = 1  #Extremity
        node_topo[d==2] = 2  #Junction
        node_topo[d>2] = 3   #Bifurcation

        extremities = np.where(node_topo == 1)[0]

        for e in extremities:
            i = indices[e][0]     #there should be only 1 element for extremities
            length = 0
            prev_node = e
            nodes_tmp = []
            nodes_tmp.append(e)
            while node_topo[i] != 1 and node_topo[i] != 3: #(not(i in bifurcations) and not(i in extremities)):
                nodes_tmp.append(i)
#                idx_edge = _findIndexWithEntry(edges_array, min(prev_node, i), max(prev_node, i))[0][0]
                idx_edge = dok[min(prev_node, i), max(prev_node, i)]   #TB 200316: faster
                length += w[idx_edge]
                idx = np.where(np.asarray(indices[i]) != prev_node)[0][0]
                prev_node = i        #Previous node
                i = indices[i][idx]  #There should be 2 elements for junction so we take the node we don't come from

            #Adding the last node (bifurcation or extremity)
#            idx_edge = _findIndexWithEntry(edges_array, min(prev_node, i), max(prev_node, i))[0][0]
            idx_edge = dok[min(prev_node, i), max(prev_node, i)]   #TB 200316: faster
            length += w[idx_edge]
            nodes_tmp.append(i)
            prev_node = i

            typeOfFil = -1
            if node_topo[prev_node] == 1: #prev_node in extremities:
                typeOfFil = 1
            elif node_topo[prev_node] == 3: #prev_node in bifurcations:
                typeOfFil = 2

            #If length too small, we remove this branch
            if (length < b) and (typeOfFil == 2):
                node_to_throw = nodes_tmp[::-1][1::]
                iid = np.mod(np.argwhere(np.in1d(new_edges.ravel('F'), node_to_throw)).T[0], len(new_edges))
                new_edges = np.delete(new_edges, iid, axis=0)
                newF[node_to_throw] = 0
                nodes_to_throw.append(nodes_tmp[0:len(nodes_tmp)-1])

        G = ig.Graph(new_edges.tolist())
        w = na.compute_weights(F, new_edges)
        G.es.set_attribute_values('weight', w)

        if len(nodes_to_throw) == 0:
            break

    #The new set of nodes and edges
    F_refined = newF
    idx_toThrow = np.unique(np.where(F_refined == [0] * D)[0])
    idx_refined = np.arange(len(F))
    idx_refined = np.delete(idx_refined, idx_toThrow, axis=0)
    F_refined = np.delete(F_refined, idx_toThrow, axis=0)

    #Go along oll thrown nodes and decrease the indices consequently to get correct remaining edges
    for (inc, i) in enumerate(idx_toThrow[::-1]):
        new_edges[new_edges > i] -= 1           #TB 200317: a bit slow..

    ed_ref = []
    for e in new_edges:
        ed_ref.append((e[0], e[1]))

    #Regularized graph structure for output
    regGraph_refined = regGraph_()
    regGraph_refined.F = F_refined
    regGraph_refined.alpha = regGraph.alpha
    regGraph_refined.edges = ed_ref

    if isinstance(regGraph.pi_k, np.ndarray):
        regGraph_refined.pi_k = regGraph.pi_k[idx_refined]
    else:
        regGraph_refined.pi_k = regGraph.pi_k

    if isinstance(regGraph.sig, np.ndarray):
        regGraph_refined.sig = regGraph.sig[idx_refined]
    else:
        regGraph_refined.sig = regGraph.sig

    if verbose >= 1:
        print('graphRefinement -- {:d} iterations done in {:.3f} seconds'.format(cnt, time.time()-t0))
        print('graphRefinement -- ' + str(len(F) - len(F_refined)) + ' nodes removed during the refinement')

    return regGraph_refined



def connectivity(pos, Rmax, F, edges, display_progress_bar=True):
    '''
    Computes the connectivity of a point at a given position and considering the graph structure
    [+] pos: (N,D) numpy array
            Positions to study
    [+] Rmax: (N,) numpy array
            Maximum euclidean distance to look around the positions pos
    [+] F: (K, D) numpy array
            Nodes positions of the graph
    [+] edges: list
            Set of all doublets of indices linked in the graph

    [-] (iclos, dclos): list of 2 floats
            Index and distance of closest node to the "pos" position
    [-] life: (Nc, 3) numpy array
            Encodes all changes in the connectivity around the location with variation of the distance from it
            col1: Connectivity events (+/-)
            col2: Geodesic distance of the event from the initial node
            col3: Euclidean distance of the event from the initial node

    Note: the function returns only events that intervienes at a euclidean distance < Rmax from the initial position on the graph

    Example of plot:
    ----------------
    ii = 2   # 1: Geodesic distance   2: Euclidean distance
    idd = np.argsort(life[:, ii])
    kappa = np.cumsum(life[idd, 0])
    plt.figure();
    pltx = life[idd, ii]
    plty = kappa
    if pltx[-1] < Rmax:
        pltx = np.hstack((life[idd, ii], Rmax))
        plty = np.hstack((kappa, kappa[-1]))
    plt.step(pltx, plty, where='post')
    plt.xlabel('Geodesic distance')
    plt.ylabel('Connectivity')
    '''

    all_life = []
    all_iclos = []
    all_dclos = []

    G = ig.Graph(edges)

    indices = G.get_adjlist()    #Could be done without igraph
    d = np.asarray(G.degree())   #Could be computed without igraph
    weights = na.compute_weights(F, edges)
    G.es.set_attribute_values('weight', weights)

    kdt = cKDTree(F)

    if display_progress_bar:
        pbar = tqdm(total=len(pos), initial=0, position=0, leave=True)

    for (i,p) in enumerate(pos):
        nodes_done = []
        next_nodes = []
        life = []

        #Find the closest node in F from position pos
        dclos, iclos = kdt.query(p)

        nodes_done.extend([iclos])
        next_nodes.extend(indices[iclos])

        #Initially, the connectivity is just the node degree
        life.append((d[iclos], 0, 0))   #death/birth, Dg, De

        while len(next_nodes) > 0:
            n = next_nodes[0]

            de = np.linalg.norm(F[iclos] - F[n])     #Euclidean distance from starting point

            dg = np.sum(weights[G.get_shortest_paths(iclos, n, output='epath')])   #Geodesic distance from starting point

            if de < Rmax[i] and d[n] == 1:   #Extremity kills a connectivity (Record the event only if de < R)
                life.append((-1, dg, de))
            elif de < Rmax[i] and d[n] > 2:  #Bifurcations creates connectivity (Record the event only if de < R)
                life.append((d[n]-2, dg, de))

            #Check next nodes to process and add them to path if still in bound
            nodes_done.extend([n])
            idxToAdd = indices[n]
            idxToAdd = np.delete(idxToAdd, np.argwhere(np.in1d(idxToAdd, nodes_done)).T[0])    #If the node is already done

            ll = np.argwhere(np.in1d(idxToAdd, next_nodes))
            if ll.shape[0] > 0:
                idxToAdd = np.delete(idxToAdd, ll[0])    #If node is already in the list

            if de < Rmax[i]:
                next_nodes.extend(idxToAdd)

            del next_nodes[0]

        life = np.vstack(life)
        idd = np.argsort(life[:,2])     #Sort events by increasing euclidean distance
        life = life[idd, :]

        all_life.append(life)
        all_dclos.append(dclos)
        all_iclos.append(iclos)

        if display_progress_bar:
            pbar.update(1)

    if display_progress_bar:
        pbar.close()

    all_life = np.array(all_life)
    all_dclos = np.array(all_dclos)
    all_iclos = np.array(all_iclos)

    return all_iclos, all_dclos, all_life



def radial_distance_skeleton(X, F, edges):

    D = X.shape[1]

    #for each position x, project it on the filamnetary structure
    pbar = tqdm(total=len(X), initial=0, position=0, leave=True)
    residuals = np.zeros(len(X))

    arr_edges = np.array(edges)  #From list to numpy array to allow indexing

    nn = NearestNeighbors(n_neighbors=10).fit(F)

    for (index, p) in enumerate(X):

        #Restrict to only a limited number of edges
        knn = nn.kneighbors(np.array([p]))[1][0]
        idx1 = np.argwhere(np.in1d(arr_edges[:,0], knn)).T[0]
        idx2 = np.argwhere(np.in1d(arr_edges[:,1], knn)).T[0]
        idx = np.unique(np.hstack((idx1, idx2)))

        edg = arr_edges[idx,:].T

        if D == 2:
                dist, posNearest = utility.vectorized_pnt2line(np.hstack((p, 0)),
                                                               np.hstack((F[edg[0]], np.array([np.zeros(len(edg[1]))]).T)),
                                                np.hstack((F[edg[1]], np.array([np.zeros(len(edg[1]))]).T)))
        elif D == 3:
                dist, posNearest = utility.vectorized_pnt2line(p, F[edg[0]], F[edg[1]])

        dist = np.array(dist)
        posNearest = np.array(posNearest)

        argmin = np.argmin(dist)
        d = dist[argmin]

        residuals[index] = d

        pbar.update(1)

    pbar.close()

    return residuals



#==============================================================================
#       READ, WRITE AND CONVERT FILAMENTS
#==============================================================================
def convertToDataFrame(filaments):
    '''
    Convert a list of filaments into a pandas DataFrame

    [+] filaments: list
            Set of filaments_ class instances

    [-] df: pandas DataFrame
            Contains the same information as filament_ class but in a DataFrame format
    '''

    fil_nodes = []
    fil_edges = []
    fil_length = []
    fil_curvature = []
    fil_residuals = []
    fil_datapoints = []
    fil_type = []
    fil_uncertainty = []

    for f in filaments:
        fil_length.append(f.length)
        fil_curvature.append(f.curvature)
        fil_type.append(f.type)
        fil_nodes.append(f.nodes)
        fil_edges.append(f.edges)
        fil_residuals.append(f.residuals)
        fil_datapoints.append(f.datapoints)
        fil_uncertainty.append(f.spatial_uncertainty)

    fil_nodes = np.array(fil_nodes)
    fil_edges = np.array(fil_edges)
    fil_length = np.array(fil_length)
    fil_curvature = np.array(fil_curvature)
    fil_residuals = np.array(fil_residuals)
    fil_datapoints = np.array(fil_datapoints)
    fil_type = np.array(fil_type)
    fil_uncertainty = np.array(fil_uncertainty)

    df = pd.DataFrame()
    df.insert(0, 'length', fil_length)
    df.insert(1, 'curvature', fil_curvature)
    df.insert(2, 'type', fil_type)
    df.insert(3, 'nodes', fil_nodes)
    df.insert(4, 'edges', fil_edges)
    df.insert(5, 'residuals', fil_residuals)
    df.insert(6, 'assignment', fil_datapoints)
    df.insert(7, 'uncertainty', fil_uncertainty)

    return df



def convertToFilaments(df):
    '''
    Convert a pandas DataFrame into a list of filaments instances

    [+] df: pandas DataFrame
            Contains the same information as filaments_ class but in a DataFrame format

    [-] filaments: list
            Set of filaments_ class instances
    '''

    filaments = []

    for i in range(len(df)):
        filament = filaments_()
        filament.length = df['length'][i]
        filament.curvature = df['curvature'][i]
        filament.type = df['type'][i]
        filament.nodes = df['nodes'][i]
        filament.edges = df['edges'][i]
        filament.residuals = df['residuals'][i]
        filament.datapoints = df['assignment'][i]
        filament.uncertainty = df['uncertainty'][i]
        filaments.append(filament)

    return filaments



def uncertaintyBalls(data, F, param, display_progress_bar):
    '''
    Compute uncertainty associated to each points in F using a bootstrap approach

    [+] data: NxD numpy array
            Array of D-dimensional datapoints
    [+] F: KxD numpy array
            Positions of reference graph nodes
    [+] param: Instance of 'param_' class
            The set of parameters to use
    [+] display_progress_bar=0: float
            If set to 1, a progress bar is displayed

    [-] rho: Kx1 numpy array
            The mean distance from each points in F to the other param.B realizations of regularized graph
    '''

    verbose = param.verbose_level

    if verbose >= 1:
        print('uncertaintyBalls -- Starting bootstrap')

    B = param.B
    allF, allEdges = bootstrap(data, param, display_progress_bar=display_progress_bar)

#    md = []

    if verbose >= 1:
        print('')
        print('uncertaintyBalls -- Starting computing distances')

    if display_progress_bar:
        pbar = tqdm(total=param.B, initial=0, position=0, leave=True)

    rho = np.zeros(len(F))
    for b in range(B):
        #Build a cKDTree on the current graph
        kdtree_b = cKDTree(allF[b])

        #Find the minimum distance from all references in the bootstrap graph
        mindist = kdtree_b.query(F, k=1)[0]
#        md.append(mindist)
        rho += 1/B * mindist**2

        if display_progress_bar:
            pbar.update(1)

#    md = np.vstack(md)
#    rho = np.sqrt(np.mean(md, axis=0))
    rho = np.sqrt(rho)

    if display_progress_bar:
        pbar.close()

    return rho



def writeCatalog_h5(data, regGraph, filaments, param, filename):
    '''
    Write a catalog of branches in h5 format

    [+] data: NxD numpy array
            Datapoints positions
    [+] F: KxD numpy array
            Nodes position
    [+] edges: Kx2 numpy array
            Indices representing connections between nodes
    [+] filaments: list
            Regularization_GMM.filaments_ objects of detected filaments
    [+] param: Instance of 'param_' class
            The set of parameters to use
    [+] filename: string
            Name of the h5 file (followed by extension .hdf5)
    '''

    df = convertToDataFrame(filaments)

    #Create h5 file
    file = h5py.File(filename, "w")

    #Parameters
    file.attrs[u'file_name'] = filename
    file.attrs[u'N_datapoints'] = len(data)
    file.attrs[u'N_nodes'] = len(regGraph.F)
    file.attrs[u'N_edges'] = len(regGraph.edges)

    params = file.create_group("param")
    params.attrs[u'lambda'] = param.lam
    params.attrs[u'pruning_level'] = param.denoisingParameter
    params.attrs[u'max_iter'] = param.maxIter
    params.attrs[u'A0'] = param.A0
    params.attrs[u'variance'] = param.sig
    params.attrs[u'topology_function'] = param.topology.__name__
    params.attrs[u'verbose_level'] = param.verbose_level
    params.attrs[u'B'] = param.B
    params.attrs[u'perc'] = param.perc
    params.attrs[u'minSigma'] = param.minSigma
    params.attrs[u'background_noise'] = param.background_noise
    params.attrs[u'domain_area'] = param.domain_area
    params.attrs[u'alpha'] = param.alpha
    params.attrs[u'covariance_update'] = param.covariance_update

    #Datapoints and graph data
    file.create_dataset("datapoints_pos", data=data)
#    file.create_dataset("nodes_pos", data=F)
#    file.create_dataset("edges", data=edges)

    graph = file.create_group("graph")
    graph.create_dataset("nodes_pos", data=regGraph.F)
    graph.create_dataset("edges", data=regGraph.edges)
    graph.create_dataset("pi_k", data=regGraph.pi_k)
    graph.create_dataset("variances", data=regGraph.sig)
    graph.create_dataset("alpha", data=regGraph.alpha)


    #Filaments group
    Nfil = len(filaments)
    fil = file.create_group("filaments")
    fil.attrs[u'N_filaments'] = Nfil

    if Nfil > 0:
        dt_float = h5py.special_dtype(vlen=np.dtype('float64'))
        dt_int = h5py.special_dtype(vlen=np.dtype('int'))

        fil.create_dataset("length", data=df['length'])
        fil.create_dataset("curvature", data=df['curvature'])
        fil.create_dataset("type", data=df['type'])

        fil.create_dataset("nodes", (len(filaments),), dtype=dt_int)
        fil["nodes"][...] = df['nodes']

        fil.create_dataset("edges", (len(filaments),), dtype=dt_int)
        fil["edges"][...] = df['edges']

        fil.create_dataset("residuals", (len(filaments),), dtype=dt_float)
        fil["residuals"][...] = df['residuals']

        fil.create_dataset("assignment", (len(filaments),), dtype=dt_int)
        fil["assignment"][...] = df['assignment']

        fil.create_dataset("uncertainty", (len(filaments),), dtype=dt_float)
        fil["uncertainty"][...] = df['uncertainty']

    file.close()

    return



def readCatalog_h5(filename):
    '''
    Read a catalog of branches in h5 format

    [+] filename: string
            Name of the h5 file (followed by extension .hdf5)

    [-] data: NxD numpy array
            Datapoints positions
    [-] F: KxD numpy array
            Nodes position
    [-] edges: Kx2 numpy array
            Indices representing connections between nodes
    [-] filaments: list
            Regularization_GMM.filaments_ objects of detected filaments
    '''

    file = h5py.File(filename, "r")
    t0 = time.time()

    data = np.array(file["datapoints_pos"])

    #Reconstruct regGraph object
    regGraph = regGraph_()
    regGraph.F = np.array(file['graph']["nodes_pos"])
    regGraph.edges = np.array(file['graph']["edges"])
    regGraph.pi_k = np.array(file['graph']['pi_k'])
    if regGraph.pi_k.shape == ():
        regGraph.pi_k = float(regGraph.pi_k)
    regGraph.alpha = float(np.array(file['graph']['alpha']))
    regGraph.sig = np.array(file['graph']['variances'])
    if regGraph.sig.shape == ():
        regGraph.sig = float(regGraph.sig)

    Nfil = file['filaments'].attrs[u'N_filaments']

    if Nfil > 0:
        filaments = pd.DataFrame()
        filaments.insert(0, 'length', np.array(file['filaments']['length']))
        filaments.insert(1, 'curvature', np.array(file['filaments']['curvature']))
        filaments.insert(2, 'type', np.array(file['filaments']['type']))
        filaments.insert(3, 'nodes', np.array(file['filaments']['nodes']))
        filaments.insert(4, 'edges', np.array(file['filaments']['edges']))
        filaments.insert(5, 'residuals', np.array(file['filaments']['residuals']))
        filaments.insert(6, 'assignment', np.array(file['filaments']['assignment']))
        filaments.insert(7, 'uncertainty', np.array(file['filaments']['uncertainty']))
    else:
        filaments = []

    param = param_()
    param.lam = file['param'].attrs['lambda']
    param.denoisingParameter = file['param'].attrs['pruning_level']
    param.maxIter = file['param'].attrs['max_iter']
    param.A0 = file['param'].attrs['A0']
    param.sig = file['param'].attrs['variance']
    param.topology = file['param'].attrs['topology_function']
    param.verbose_level = file['param'].attrs['verbose_level']
    param.B = file['param'].attrs['B']
    param.perc = file['param'].attrs['perc']
    param.alpha = file['param'].attrs['alpha']
    param.background_noise = file['param'].attrs['background_noise']
    param.domain_area = file['param'].attrs['domain_area']
    param.minSigma = file['param'].attrs['minSigma']
    param.covariance_update = file['param'].attrs['covariance_update']

    tr = time.time() - t0

    print('File read in {:.4f} seconds with {:d} datapoints and {:d} filaments'.format(tr, len(data), len(filaments)))
    print()
    print('Parameters used:')
    print('-----------------')
    print('lambda = {:.1f}'.format(param.lam))
    print('denoising = {:.1f}'.format(param.denoisingParameter))
    print('variance = {:.3f}'.format(param.sig))
    print('topology = {}'.format(param.topology))
    print('Covariance = {}'.format(param.covariance_update))
    print('Uniform background = {}'.format(param.background_noise))

    return data, regGraph, filaments, param



import numpy as np
import scipy.stats as st
from tqdm import tqdm
import scipy.sparse as sp


def _computeResponsability(x, F, sigs):
    '''
    -- INTERNAL FUNCTION --
    Compute the responsability of a datapoint x given a set of centroids F
    WARNING: Assumes an isotropic diagonal covariance for each cluster K (sigs is a Kx1 numpy array)
    '''

    dst = np.linalg.norm(x - F, axis=1)**2
    pr = np.exp(-dst / 2 / sigs)
    resp = pr / np.sum(pr)
    return resp


def _mvn(mu, sig, X):
    '''
    -- INTERNAL FUNCTION --
    Gaussian pdf computed of a position X for several gaussian clusters.
    The gaussian is assumed to be spherical with scalar covariance.

    [+] mu: (K,D) numpy array
            Mean of gaussian clusters
    [+] sig: (K,) numpy array
            Variance of gaussian clusters
            If scalar, assumes the same for all clusters
    [+] X (D,) numpy array
            Contains position to evaluate the pdf for each gaussian

    [-] pdf: (K,) numpy array
            Value of the gaussian pdf computed at position X for all different gaussians
    '''
    D = len(X)
    A = 1 / (2*np.pi*sig)**(D/2)
    pdf = A * np.exp(-1/2 * np.linalg.norm(X-mu, axis=1)**2 / sig)
    return pdf


def _mvnX(mu, sig, X):
    '''
    -- INTERNAL FUNCTION --
    Gaussian pdf computed of a position X for several gaussian clusters.
    The gaussian is assumed to be spherical with scalar covariance.

    [+] mu: (D,) numpy array
            Mean of gaussian
    [+] sig: scalar
            Variance of gaussian clusters
    [+] X (N,D) numpy array
            Contains positions to evaluate the pdf

    [-] pdf: (N,) numpy array
            Value of the gaussian pdf computed at positions X
    '''

    D = len(mu)
    if isinstance(sig, np.ndarray):
        A = 1 / ((2*np.pi)**(D/2)*np.linalg.det(sig)**(1/2))
        invSig = np.linalg.inv(sig)
        pdf = A * np.exp(-1/2 * (X-mu).T @ invSig @ (X-mu))
    else:
        A = 1 / (2*np.pi*sig)**(D/2)
        pdf = A * np.exp(-1/2 * np.linalg.norm(X-mu, axis=1)**2  / sig)

    return pdf


def _normalize(R, cte=0):
    '''
    -- INTERNAL FUNCTION --
    Normalize a scipy sparse matrix over rows

    [+] R: scipy sparse matrix
            The matrix to be normalized
    [+] cte: scalar
            Constant to add to the normalization constant
    '''
    R = R.tocsr()

    norm = cte + np.repeat(np.add.reduceat(R.data, R.indptr[:-1]), np.diff(R.indptr))
    R.data = R.data / norm
    return R, norm


def updatePositionsGMM(data, R, A):
    '''
    Update positions of centers of gaussian clusters by solving linear system in each dimension.

    [+] data: NxD numpy array
            Array of D-dimensional datapoints
    [+] R: NxK numpy array
            Responsability matrix (support scipy sparse matrix)
    [+] A: KxK numpy array
            Matrix A = lambda*L + Lambda (support scipy sparse matrix)

    [-] F: KxD numpy array
            Updated clusters means positions
    '''

    K = R.shape[1]
    D = len(data.T)

    #Solve linear system for each dimension
    lu_obj = sp.linalg.splu(A)
    XR = data.T @ R
    F = np.zeros((K, D))
    for j in range(D):
        F[:,j] = lu_obj.solve(XR[j,:])
    return F


#import time
def M_step(data, V, R, A, Lambda, update_pos_only=True, background_noise=False, resp_background=[]):
    '''
    Compute the M-step of the regularized GMM problem

    [+] data: NxD numpy array
            Array of D-dimensional datapoints
    [+] R: NxK numpy array
            Responsability matrix (support scipy sparse matrix)
    [+] A: KxK numpy array
            Matrix A = lambda*L + Lambda (support scipy sparse matrix)
    [+] Lambda: KxK sparse diagonal matrix
            Diagonal matrix containing sum of responsabilities for each cluster
    [+] update_pos_only: boolean
            If True, update only the cluster means and leave rest unchanged
    [+] background_noise: boolean
            If True, also update alpha, the amplitude of background noise
    [+] resp_background: (N,) numpy array
            Responsabilities of the background for each datapoint

    [-] F: KxD numpy array
            Updated cluster mean positions
    [-] variance: scalar or (K,) numpy array
            Updated cluster variance
    [-] pi_k: empty list or (K,) numpy array
            Updated amplitudes associated to each cluster
    [-] alpha: scalar
            Updated background noise amplitude
    '''

    N = data.shape[0]
    D = data.shape[1]
    pi_k = []
    variance = 0
    alpha = 0

    #Position update
#    t0 = time.time()
    F = updatePositionsGMM(data, V, A)

    if not update_pos_only:
        #Variance update
        nk = Lambda.data[0]
#        t1 = time.time()
        if D == 2:
            diagCov, antiDiag = updateCov2D(data, F, R, nk, diag_only=True)
        elif D == 3:
            diagCov, covXY, covXZ, covYZ = updateCov3D(data, F, R, nk, diag_only=True)
        variance = np.mean(diagCov, axis=1)
#        t2 = time.time()

        #Amplitudes update
        pi_k = 1/N * nk

        #Background amplitude update
        alpha = 0
        if background_noise:
            alpha = 1/N * np.sum(resp_background)

#        t3 = time.time()

#        print('Time position : {:.3f}'.format(t1-t0))
#        print('Time variance : {:.3f}'.format(t2-t1))
#        print('Time amplitudes : {:.3f}'.format(t3-t2))
    return F, variance, pi_k, alpha



def evaluateObjective(data, F, R, Lambda, L, lam, sigs, pi_k, full=1, background_noise=False, alpha=0, domain_area=0):
    ''' Evaluate the objective function computed as the lower bound of the EM algorithm
    Basically, it corresponds to the sum of 3 terms: Q, the expectation function, H the log entropy and the prior log(p(\theta))

    [+] data: NxD numpy array
            Array of D-dimensional datapoints
    [+] F: KxD numpy array
            Clusters means positions
    [+] R: NxK numpy array
            Responsability matrix (support sparse matrix)
    [+] Lambda: KxK diagonal sparse matrix
            The expected number of datapoints in each clusters (Lambda.data=R.sum(axis=0))
    [+] L: KxK sparse matrix
            Laplacian matrix of the current graph structure
    [+] lam: float > 0
            Regularization parameter
    [+] sigs: Kx1 numpy array
            Current estimate of variance for each spherical cluster

    [-] obj: float
            Estimated value of the objective function
    '''
    if full:
        log_ptheta = - lam*np.trace(F.T@L@F)

        #Greedy computation of the log-likelihood
#        Loglikelihood = 0
#        for i in range(len(data)):
#            pdf = _mvn(F, sigs, data[i])
#            if background_noise:
#                Loglikelihood += np.log(np.sum(pi_k*pdf) + alpha/domain_area)
#            else:
#                Loglikelihood += np.log(np.sum(pi_k*pdf))

        ss = np.ones(len(F))*sigs

        #Compute the full log-likelihood
        D = data.shape[1]
        log_prob = (np.sum((F ** 2), axis=1)*(1/ss) - 2. * np.dot(data, (F.T*(1/ss))) + np.outer(np.linalg.norm(data, axis=1)**2, (1/ss).T))
        log_det = D * np.log(1/ss**0.5)
        ll = -.5 * (data.shape[1] * np.log(2 * np.pi) + log_prob) + log_det

        probs = ll+np.log(pi_k)
        if background_noise:
            log_prob_norm = np.log(np.sum(np.exp(probs)+alpha/domain_area, axis=1))
        else:
            log_prob_norm = np.log(np.sum(np.exp(probs), axis=1))
        loglikelihood = np.sum(log_prob_norm)

        regLog = loglikelihood + log_ptheta
        lower_bound = regLog
#        print('LogL = '+ str(loglikelihood))
#        print('Prior= '+ str(log_ptheta))
#        print('===========================')

    else:
#        Q = R.data*np.log(pi_k) - np.trace(data.T@data) - np.trace(F.T@Lambda@F) + 2*np.trace(data.T@R@F)  #Expectation term
        log_ptheta = - lam*np.trace(F.T@L@F)
##        H = sigs**2*np.sum(R.data*np.log(R.data))        #entropy term
#        lower_bound = Q + log_ptheta #- H  #Q + log(P(theta)) - H is the full posterior


        #2nd attempt
#        ksi = sp.diags(np.array(R.sum(0))[0])
        D = data.shape[1]

#        ksi = sp.diags(np.array(R.sum(1)).T[0])
#        Q = np.sum(R.multiply(sigs)*np.log(pi_k)) - D/2*np.sum(R.multiply(sigs)*np.log(2*np.pi)) - D/2*np.sum(R.multiply(np.log(sigs)))
#        Q = -1/2*(np.trace(data.T@ksi@data) + np.trace(F.T@Lambda@F) - 2*np.trace(data.T@R@F))  #what is inside the exponential
#        lower_bound = Q

        #Compute the full log-likelihood (thanks to scikit-learn)
        log_prob = (np.sum((F ** 2), axis=1)*(1/sigs) - 2. * np.dot(data, (F.T*(1/sigs))) + np.outer(np.linalg.norm(data, axis=1)**2, (1/sigs).T))
        log_det = D * np.log(1/sigs**0.5)
        ll = -.5 * (data.shape[1] * np.log(2 * np.pi) + log_prob) + log_det
        log_prob = np.log(np.sum(np.exp(ll+np.log(pi_k)), axis=1))
        loglikelihood = np.sum(log_prob)

        lower_bound = loglikelihood + log_ptheta


    return lower_bound


def evaluateLL(data, F, sigs, pi_k):
    D = data.shape[1]
    log_prob = (np.sum((F ** 2), axis=1)*(1/sigs) - 2. * np.dot(data, (F.T*(1/sigs))) + np.outer(np.linalg.norm(data, axis=1)**2, (1/sigs).T))
    log_det = D * np.log(1/sigs**0.5)
    ll = -.5 * (data.shape[1] * np.log(2 * np.pi) + log_prob) + log_det
    log_prob_norm = np.log(np.sum(np.exp(ll+np.log(pi_k)), axis=1))
    loglikelihood = np.sum(log_prob_norm)

    return loglikelihood


def evaluateObjective_local(data, F, R, Lambda, L, lam, sigs, pi_k, full=1, background_noise=False, alpha=0, domain_area=0, grid=[]):

#    gridx = grid[0]
#    gridy = grid[1]

#    map_obj = np.zeros((64, 64))
    map_obj = np.zeros(len(F))

    h = 1 #0.2

    if full:
#        log_ptheta = - lam*np.trace(F.T@L@F)
        all_pdf = []

        for i in range(len(data)):
            pdf = _mvn(F, sigs, data[i])
            all_pdf.append(np.log(np.sum(pi_k*pdf)))

        all_pdf = np.array(all_pdf)

        #Evaluate on the grid
#        for (ix, xx) in enumerate(gridx):
#            for (iy, yy) in enumerate(gridy):
#                Loglikelihood = 0
#                weights = _mvnX(np.array([xx, yy]), h**2, data)
#                Loglikelihood += np.sum(weights * all_pdf)
#
#                map_obj[ix, iy] = Loglikelihood #+ log_ptheta

        #Evaluate at nodes positions
        for (ii, ff) in enumerate(F):
            Loglikelihood = 0
            weights = _mvnX(ff, h**2, data)
            Loglikelihood += np.sum(weights*all_pdf)

            map_obj[ii] = Loglikelihood

    return map_obj


def neighborsCov(kdtree, data, centers, k):
    D = len(data.T)
    dst, nbrs = kdtree.query(centers, k=k)
    x = data[nbrs]

    covs = []
    cov00 = []
    cov11 = []
    cov01 = []

    cov02 = []
    cov12 = []
    cov22 = []

    for x in data[nbrs]:
        cov = np.cov(x.T)
        covs.append(cov)
        cov00.append(cov[0,0])
        cov11.append(cov[1,1])
        cov01.append(cov[0,1])

        if D == 3:  #In 3D, add other terms
            cov02.append(cov[0,2])
            cov12.append(cov[1,2])
            cov22.append(cov[2,2])

    cov00 = np.array(cov00)
    cov11 = np.array(cov11)
    cov22 = np.array(cov22)
    cov01 = np.array(cov01)
    cov02 = np.array(cov02)
    cov12 = np.array(cov12)

    return cov00, cov11, cov22, cov01, cov02, cov12, covs


def diagCov(data, F, R, nk):
    ''' Estimate the diagonal elements of the covariance matrices

    [+] data: NxD numpy array
            Array of D-dimensional datapoints
    [+] F: KxD numpy array
            Clusters means positions
    [+] R: NxK numpy array
            Responsability matrix (support sparse matrix)
    [+] nk: Kx1 numpy array
            Current number of data associated to kernels (=R.sum(axis=0))

    [-] diagCov: KxD numpy array
            Diagonal elements of all covariance matrices

    Note: For example, in 2D, the covariance matrix of cluster k is then C[k] = [[diagCov[k, 0], 0], [0, diagCov[k, 1]]]
    '''
    avg_X2 = R.T@data**2 / nk[:, np.newaxis]
    avg_F2 = F**2
    avg_XF = R.T@data / nk[:, np.newaxis] * F
    diagCov = avg_X2 - 2 * avg_XF + avg_F2
    return diagCov



def updateCov2D(data, F, R, nk, diag_only=0):
    ''' Estimate all elements of all covariance matrices in the 2D case

    [+] data: NxD numpy array
            Array of D-dimensional datapoints
    [+] F: KxD numpy array
            Clusters means positions
    [+] R: NxK numpy array (support scipy sparse matrix)
            Responsability matrix
    [+] nk: Kx1 numpy array
            Current number of data associated to kernels (=R.sum(axis=0))
    [+] diag_only: Boolean
            To only return the diagonal elements of the matrix

    [-] diagCov: Kx2 numpy array
            Diagonal elements of all covariance matrices
    [-] antiDiag: Kx2 numpy array
            Antidiagonal elements of all covariance matrices

    Note: The full covariance matrix of cluster k is then C[k] = [[diagCov[k,0], antiDiag[k]], [antiDiag[k], diagCov[k,1]]]
    '''
    diag = diagCov(data, F, R, nk)

    antiDiag = []
    if diag_only == False:
        #Antidiagonal terms
        avg_Xxy = R.T @ (data.T[0] * data.T[1]) / nk
        avg_Fxy = (F.T[0] * F.T[1])
        avg_XyFx = (R.T @ data).T[1] * F.T[0] / nk
        avg_XxFy = (R.T @ data).T[0] * F.T[1] / nk
        antiDiag = avg_Xxy - avg_XyFx - avg_XxFy + avg_Fxy

    return diag, antiDiag



def updateCov3D(data, F, R, nk, diag_only=False):
    ''' Estimate all elements of all covariance matrices in the 3D case
    The full symmetric covariance matrix can be retrieved as:
            [[diag[0], xy, xz],
             [xy, diag[1], yz],
             [xz, yz, diag[2]]]

    [+] data: NxD numpy array
            Array of D-dimensional datapoints
    [+] F: KxD numpy array
            Clusters means positions
    [+] R: NxK numpy array (support scipy sparse matrix)
            Responsability matrix
    [+] nk: Kx1 numpy array
            Current number of data associated to kernels (=R.sum(axis=0))
    [+] diag_only: Boolean
            To only return the diagonal elements of the matrix

    [-] diagCov: Kx2 numpy array
            Diagonal elements of all covariance matrices
    [-] xy: Kx1 numpy array
            XY covariances
    [-] xz: Kx1 numpy array
            XZ covariances
    [-] yz: Kx1 numpy array
            YZ covariances
    '''
    #Diagonal elements
    diag = diagCov(data, F, R, nk)

    xy = []
    xz = []
    yz = []
    if diag_only == False:
        #XY
        avg_Xxy = R.T @ (data.T[0] * data.T[1]) / nk
        avg_Fxy = (F.T[0] * F.T[1])
        avg_XyFx = (R.T @ data).T[1] * F.T[0] / nk
        avg_XxFy = (R.T @ data).T[0] * F.T[1] / nk
        xy = avg_Xxy - avg_XyFx - avg_XxFy + avg_Fxy

        #XZ
        avg_Xxz = R.T @ (data.T[0] * data.T[2]) / nk
        avg_Fxz = (F.T[0] * F.T[2])
        avg_XzFx = (R.T @ data).T[2] * F.T[0] / nk
        avg_XxFz = (R.T @ data).T[0] * F.T[2] / nk
        xz = avg_Xxz - avg_XzFx - avg_XxFz + avg_Fxz

        #YZ
        avg_Xzy = R.T @ (data.T[2] * data.T[1]) / nk
        avg_Fzy = (F.T[2] * F.T[1])
        avg_XyFz = (R.T @ data).T[1] * F.T[2] / nk
        avg_XzFy = (R.T @ data).T[2] * F.T[1] / nk
        yz = avg_Xzy - avg_XyFz - avg_XzFy + avg_Fzy
    return diag, xy, xz, yz



def eigvals2D(a, b, c):
    ''' Compute all eigenvalues of K 2x2 symmetric matrix
    This implementation is faster than using np.linalg.eigvals on each matrix

    [+] a: Kx1 numpy array
            K first diagonal elements of K matrices
    [+] b: Kx1 numpy array
           K second diagonal elements of K matrices
    [+] c: Kx1 numpy array
            K antidiagonal elements of K matrices

    [-] eigs: 2x1 numpy array
            Eigenvalues in decreasing order
    '''
    trace = a + b
    det = a*b - c*c

    l1 = trace/2 + np.sqrt(trace**2/4 - det)
    l2 = trace/2 - np.sqrt(trace**2/4 - det)

    eigs = np.vstack((l1, l2)).T

    return eigs



def eigvals3D(a, b, c, d, e, f):
    ''' Estimate all eigenvalues of K 3x3 symmetric matrix
    This implementation is faster than using np.linalg.eigvals on each matrix

    [+] a: Kx1 numpy array
            K first diagonal elements of K matrices
    [+] b: Kx1 numpy array
            K second diagonal elements of K matrices
    [+] c: Kx1 numpy array
            K third diagonal elements of K matrices
    [+] d: Kx1 numpy array
            K elements [1, 2] of K matrices
    [+] e: Kx1 numpy array
            K elements [1, 3] of K matrices
    [+] f: Kx1 numpy array
            K elements [2, 3] of K matrices

    [-] eigs: 3x1 numpy array
            Eigenvalues in decreasing order
    '''
    q = (a + b + c)/3   #Trace/3
    q2 = (a-q)**2 + (b-q)**2 + (c-q)**2 + 2*d**2 + 2*e**2 + 2*f**2   #Trace of (A-q*I)^2
    p = np.sqrt(q2 / 6)
    detB = 1/p**3*((a-q)*(b-q)*(c-q) + 2*d*f*e - e**2*(b-q) - d**2*(c-q) - f**2*(a-q))
    r = detB / 2

    phi = np.ones(len(r))*np.arccos(r)/3
    phi[r<=-1] = np.pi /3
    phi[r>=1] = 0

    #Eigenvalues in decreasing order
    eig1 = q + 2 * p * np.cos(phi)
    eig3 = q + 2 * p * np.cos(phi + (2*np.pi/3))
    eig2 = 3 * q - eig1 - eig3

    eigs = np.vstack((eig1, eig2, eig3)).T

    #    eigs[eigs<0] = 0

    return eigs



def computeDensity(x, mu, S, alpha=0):
    '''
    Compute the density on x positions from a mixture model by assuming
        a Gaussian Mixture with centers mu, covariances S and weights alpha

    [+] x: positions to compute the density on (Nxd)
    [+] mu: positions of centroids (Kxd)
    [+] S: covariance of each component (Kx(dxd) or scalar hence assuming S*I(d,d) for all clusters)
    [+] alpha: mixtures' weights (if 0, assumes equidistributed weights)

    [-] density_norm: Value of the density at each grid position
    '''

    N, d = x.shape

#    d = len(x)
#    if d == 2:
#        N = x[0].shape[0]
#        M = x[1].shape[0]
#    elif d == 3:
#        N, M, L = x[0].shape

    K = len(mu)

    try:
        if len(S.shape) == 0:
            scalar_cov = 1
        else:
            scalar_cov = 0
    except:
        scalar_cov = 1

    try:
        len(alpha)
        equidistributed_weights = 0
    except:
        equidistributed_weights = 1
        alpha = 1/K

#    if d == 2:
#        xx, yy = x
#        pos = np.empty(xx.shape + (2,))
#        pos[:, :, 0] = xx; pos[:, :, 1] = yy
#        density = np.zeros((N, M))
#    elif d == 3:
#        xx, yy, zz = x
#        pos = np.empty(xx.shape + (3,))
#        pos[:, :, :, 0] = xx; pos[:, :, :, 1] = yy; pos[:, :, :, 2] == zz;
#        density = np.zeros((N, M, L))

    pos = x
    density = np.zeros(N)

    components = []
    pbar = tqdm(total=K, initial=0, position=0, leave=True)
    for k in range(K):
        if scalar_cov:
            components.append(st.multivariate_normal(mu[k], S*np.eye(d)))
        else:
            components.append(st.multivariate_normal(mu[k], S[k]))

        if equidistributed_weights:
            density += alpha*components[k].pdf(pos)
        else:
            density += alpha[k]*components[k].pdf(pos)
        pbar.update(1)

    pbar.close()

    #Normalization step
    density_norm = density / np.sum(density) * len(mu)

    return density_norm




import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

pkgs = setuptools.find_packages(where="src")

setuptools.setup(
    name='trex_python',
    version='0.1.2',    
    description='A Python implementation of the T-ReX algorithm',
    url='https://git.ias.u-psud.fr/tbonnair/t-rex.git',
    author='Tony Bonnaire',
    author_email='bonnaire.tonyim@gmail.com',
    license='GNU General Public License v3.0',
    package_dir={"": "src"},
    packages=pkgs,
    python_requires=">=3.7",
)